# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
from MDAnalysisTests.datafiles import PDBQT_input, PDBQT_querypdb
from MDAnalysis.lib.NeighborSearch import AtomNeighborSearch

from numpy.testing import (
    assert_equal,
)
import pytest

from MDAnalysisTests import make_Universe


class TestPDBQT(object):
    @pytest.fixture()
    def universe(self):
        """Set up the standard AdK system in implicit solvent."""
        return mda.Universe(PDBQT_input)

    def test_segid(self, universe):
        sel = universe.select_atoms('segid A')
        assert_equal(sel.n_atoms, 909, "failed to select segment A")
        sel = universe.select_atoms('segid B')
        assert_equal(sel.n_atoms, 896, "failed to select segment B")

    def test_protein(self, universe):
        sel = universe.select_atoms('protein')
        assert_equal(sel.n_atoms, 1805, "failed to select protein")
        assert_equal(sel.atoms.ix, universe.atoms.ix,
                     "selected protein is not the same as auto-generated protein segment A+B")

    def test_backbone(self, universe):
        sel = universe.select_atoms('backbone')
        assert_equal(sel.n_atoms, 796)

    def test_neighborhood(self, universe):
        '''test KDTree-based distance search around query atoms

        Creates a KDTree of the protein and uses the coordinates of
        the atoms in the query pdb to create a list of protein
        residues within 4.0A of the query atoms.
        '''
        query_universe = mda.Universe(PDBQT_querypdb)  # PDB file

        protein = universe.select_atoms("protein")
        ns_protein = AtomNeighborSearch(protein)
        query_atoms = query_universe.atoms
        residue_neighbors = ns_protein.search(query_atoms, 4.0)
        assert_equal(len(residue_neighbors), 80)

    def test_n_frames(self, universe):
        assert_equal(universe.trajectory.n_frames, 1,
                     "wrong number of frames in pdb")

    def test_time(self, universe):
        assert_equal(universe.trajectory.time, 0.0, "wrong time of the frame")

    def test_frame(self, universe):
        assert_equal(universe.trajectory.frame, 0,
                     "wrong frame number (0-based, should be 0 for single frame readers)")


class TestPDBQTWriter(object):
    reqd_attributes = ['names', 'types', 'resids', 'resnames', 'radii',
                       'charges']

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir) + 'out.pdbqt'

    def test_roundtrip_writing_coords(self, outfile):
        u = mda.Universe(PDBQT_input)
        u.atoms.write(outfile)
        u2 = mda.Universe(outfile)

        assert_equal(u2.atoms.positions, u.atoms.positions,
                     "Round trip does not preserve coordinates")

    def test_roundtrip_formatting(self, outfile):
        # Compare formatting of first line
        u = mda.Universe(PDBQT_input)
        u.atoms.write(outfile)

        with open(PDBQT_input, 'r') as inf:
            l_ref = inf.readline().strip()
        with open(outfile, 'r') as inf:
            inf.readline()  # header
            inf.readline()  # cryst
            l_new = inf.readline().strip()
        assert l_ref == l_new

    @staticmethod
    def assert_writing_warns(u, outfile):
        with pytest.warns(UserWarning):
            u.atoms.write(outfile)

    def test_write_no_charges(self, outfile):
        attrs = self.reqd_attributes
        attrs.remove('charges')
        u = make_Universe(attrs, trajectory=True)

        self.assert_writing_warns(u, outfile)

        u2 = mda.Universe(outfile)

        assert all(u2.atoms.charges == 0.0)

    def test_write_no_chainids_with_segids(self, outfile):
        attrs = self.reqd_attributes
        attrs.append('segids')
        u = make_Universe(attrs, trajectory=True)

        u.atoms.write(outfile)
        u2 = mda.Universe(outfile)

        # Should have used last letter of segid as chainid
        assert all(u2.atoms[:25].segids == 'A')
        assert all(u2.atoms[25:50].segids == 'B')

    def test_get_writer(self, outfile):
        u = mda.Universe(PDBQT_input)
        w = u.trajectory.Writer(outfile)

        assert isinstance(w, mda.coordinates.PDBQT.PDBQTWriter)
        w.close()



"""Tests for MDAnalysis.core.topology objects.

Should convert between indices (*ix)
Should work with both a single or an array of indices
"""
import itertools
from numpy.testing import (
    assert_equal,
)
import pytest
import numpy as np

from MDAnalysisTests import make_Universe

from MDAnalysis.core.topology import (
    Topology,
    TransTable,
    make_downshift_arrays,
)
from MDAnalysis.core import topologyattrs as ta
from MDAnalysis.core import groups
from MDAnalysis import NoDataError
import MDAnalysis


class TestTransTable(object):
    @pytest.fixture()
    def tt(self):
        Ridx = np.array([0, 0, 2, 2, 1, 1, 3, 3, 1, 2])
        Sidx = np.array([0, 1, 1, 0])
        return TransTable(10, 4, 2, Ridx, Sidx)

    def test_a2r(self, tt):
        for aix, rix in zip(
                [np.array([0, 1, 2]),
                 np.array([9, 6, 2]),
                 np.array([3, 3, 3])],
                [np.array([0, 0, 2]),
                 np.array([2, 3, 2]),
                 np.array([2, 2, 2])]
        ):
            assert_equal(tt.atoms2residues(aix), rix)

    def test_r2a_1d(self, tt):
        for rix, aix in zip(
                [[0, 1], [1, 1], [3, 1]],
                [[0, 1, 4, 5, 8], [4, 5, 8, 4, 5, 8], [6, 7, 4, 5, 8]]
        ):
            assert_equal(tt.residues2atoms_1d(rix), aix)

    def test_r2a_2d(self, tt):
        for rix, aix in zip(
                [[0, 1],
                 [1, 1],
                 [3, 1]],
                [[[0, 1], [4, 5, 8]],
                 [[4, 5, 8], [4, 5, 8]],
                 [[6, 7], [4, 5, 8]]]
        ):
            answer = tt.residues2atoms_2d(rix)
            for a1, a2 in zip(answer, aix):
                assert_equal(a1, a2)

    def test_r2s(self, tt):
        for rix, sidx in zip(
                [np.array([0, 1]),
                 np.array([2, 1, 0]),
                 np.array([1, 1, 1])],
                [np.array([0, 1]),
                 np.array([1, 1, 0]),
                 np.array([1, 1, 1])]
        ):
            assert_equal(tt.residues2segments(rix), sidx)

    def test_s2r_1d(self, tt):
        for sidx, rix in zip(
                [[0, 1],
                 [1, 0],
                 [1, 1]],
                [[0, 3, 1, 2],
                 [1, 2, 0, 3],
                 [1, 2, 1, 2]]
        ):
            assert_equal(tt.segments2residues_1d(sidx), rix)

    def test_s2r_2d(self, tt):
        for sidx, rix in zip(
                [[0, 1],
                 [1, 0],
                 [1, 1]],
                [[[0, 3], [1, 2]],
                 [[1, 2], [0, 3]],
                 [[1, 2], [1, 2]]]
        ):
            answer = tt.segments2residues_2d(sidx)
            for a1, a2 in zip(answer, rix):
                assert_equal(a1, a2)

    def test_s2a_1d(self, tt):
        for sidx, aix in zip(
                [[0, 1],
                 [1, 0],
                 [1, 1]],
                [[0, 1, 6, 7, 4, 5, 8, 2, 3, 9],
                 [4, 5, 8, 2, 3, 9, 0, 1, 6, 7],
                 [4, 5, 8, 2, 3, 9, 4, 5, 8, 2, 3, 9]],
        ):
            assert_equal(tt.segments2atoms_1d(sidx), aix)

    def test_s2a_2d(self, tt):
        for sidx, aix in zip(
                [[0, 1],
                 [1, 0],
                 [1, 1]],
                [[[0, 1, 6, 7], [4, 5, 8, 2, 3, 9]],
                 [[4, 5, 8, 2, 3, 9], [0, 1, 6, 7]],
                 [[4, 5, 8, 2, 3, 9], [4, 5, 8, 2, 3, 9]]],
        ):
            answer = tt.segments2atoms_2d(sidx)
            for a1, a2 in zip(answer, aix):
                assert_equal(a1, a2)

    def test_move_atom_simple(self, tt):
        assert_equal(tt.atoms2residues(1), 0)
        assert_equal(len(tt.residues2atoms_1d(0)), 2)
        assert_equal(len(tt.residues2atoms_1d(3)), 2)

        # move 2nd atom to 4th residue (atom1 -> res3)
        tt.move_atom(1, 3)

        assert_equal(tt.atoms2residues(1), 3)  # identity of parent changed
        assert_equal(len(tt.residues2atoms_1d(0)), 1)  # 1 fewer here
        assert_equal(len(tt.residues2atoms_1d(3)), 3)  # 1 more here

    def test_move_residue_simple(self, tt):
        assert_equal(tt.residues2segments(1), 1)
        assert_equal(len(tt.segments2residues_1d(0)), 2)
        assert_equal(len(tt.segments2residues_1d(1)), 2)

        # move 2nd residue to 1st segment (res1 -> seg0)
        tt.move_residue(1, 0)

        assert_equal(tt.residues2segments(1), 0)
        assert_equal(len(tt.segments2residues_1d(0)), 3)
        assert_equal(len(tt.segments2residues_1d(1)), 1)


class TestLevelMoves(object):
    """Tests for moving atoms/residues between residues/segments

    
    Atoms can move between residues by setting .residue with a Residue
    Residues can move between segments by setting .segment with a Segment

    Moves are performed by setting either [res/seg]indices or [res/seg]ids

    
    """

    @pytest.fixture()
    def u(self):
        return make_Universe(('resids', 'resnames', 'segids'))

    @staticmethod
    def assert_atoms_match_residue(atom, residue):
        # check that an Atom's residue level properties match the residue
        if isinstance(atom, groups.Atom):
            atom = [atom]  # can work with either Atom or AG
        if isinstance(residue, groups.Residue):
            residue = itertools.cycle((residue,))  # either R or RG or [R, R]
        for at, r in zip(atom, residue):
            assert at.resindex == r.resindex
            assert at.resid == r.resid
            assert at.resname == r.resname

    def test_move_atom(self, u):
        # move a single atom by providing a new Residue object
        at = u.atoms[0]
        source = u.residues[0]
        dest = u.residues[4]

        assert at in source.atoms
        assert not at in dest.atoms
        self.assert_atoms_match_residue(at, source)
        assert_equal(len(source.atoms), 5)
        assert_equal(len(dest.atoms), 5)

        at.residue = dest

        assert not at in source.atoms
        assert at in dest.atoms
        self.assert_atoms_match_residue(at, dest)
        assert_equal(len(source.atoms), 4)
        assert_equal(len(dest.atoms), 6)

    def test_move_atomgroup_single_residue(self, u):
        # move contents of AtomGroup into Residue object
        ag = u.atoms[[1, 3]]
        source = u.residues[0]
        dest = u.residues[4]

        for at in ag:
            assert at in source.atoms
            assert not at in dest.atoms
        self.assert_atoms_match_residue(ag, source)
        assert_equal(len(source.atoms), 5)
        assert_equal(len(dest.atoms), 5)

        ag.residues = dest

        for at in ag:
            assert not at in source.atoms
            assert at in dest.atoms
        self.assert_atoms_match_residue(ag, dest)
        assert_equal(len(source.atoms), 3)
        assert_equal(len(dest.atoms), 7)

    def test_move_atomgroup_residuegroup(self, u):
        # move contents of AtomGroup into Residue object
        ag = u.atoms[[1, 3]]
        source = u.residues[0]
        dest = u.residues[4] + u.residues[5]

        for at in ag:
            assert at in source.atoms
            assert not at in dest.atoms
        self.assert_atoms_match_residue(ag, source)
        assert_equal(len(source.atoms), 5)
        assert_equal(len(dest[0].atoms), 5)
        assert_equal(len(dest[1].atoms), 5)

        ag.residues = dest

        for at in ag:
            assert not at in source.atoms
            assert at in dest.atoms
        self.assert_atoms_match_residue(ag, dest)
        assert_equal(len(source.atoms), 3)
        assert_equal(len(dest[0].atoms), 6)
        assert_equal(len(dest[1].atoms), 6)

    def test_move_atomgroup_residue_list(self, u):
        # move contents of AtomGroup into Residue object
        ag = u.atoms[[1, 3]]
        source = u.residues[0]
        dest = [u.residues[4], u.residues[5]]

        for at, d in zip(ag, dest):
            assert at in source.atoms
            assert not at in d.atoms
        self.assert_atoms_match_residue(ag, source)
        assert_equal(len(source.atoms), 5)
        assert_equal(len(dest[0].atoms), 5)
        assert_equal(len(dest[1].atoms), 5)

        ag.residues = dest

        for at, d in zip(ag, dest):
            assert not at in source.atoms
            assert at in d.atoms
        self.assert_atoms_match_residue(ag, dest)
        assert_equal(len(source.atoms), 3)
        assert_equal(len(dest[0].atoms), 6)
        assert_equal(len(dest[1].atoms), 6)

    # Wrong size argument for these operations
    def test_move_atom_residuegroup_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.atoms[0], 'residue', u.atoms[1:3])

    def test_move_atom_residue_list_TE(self, u):
        dest = [u.residues[1], u.residues[3]]
        with pytest.raises(TypeError):
            setattr(u.atoms[0], 'residue', dest)

    def test_move_atomgroup_residuegroup_VE(self, u):
        ag = u.atoms[:2]
        dest = u.residues[5:10]

        with pytest.raises(ValueError):
            setattr(ag, 'residues', dest)

    def test_move_atomgroup_residue_list_VE(self, u):
        ag = u.atoms[:2]
        dest = [u.residues[0], u.residues[10], u.residues[15]]

        with pytest.raises(ValueError):
            setattr(ag, 'residues', dest)

    # Setting to non-Residue/ResidueGroup raises TE
    def test_move_atom_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.atoms[0], 'residue', 14)

    def test_move_atomgroup_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.atoms[:5], 'residues', 15)

    def test_move_atomgroup_list_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.atoms[:5], 'residues', [14, 12])

    # Test illegal moves - Atom.segment can't be changed
    def test_move_atom_segment_NIE(self, u):
        with pytest.raises(NotImplementedError):
            setattr(u.atoms[0], 'segment', u.segments[1])

    def test_move_atomgroup_segment_NIE(self, u):
        with pytest.raises(NotImplementedError):
            setattr(u.atoms[:3], 'segments', u.segments[1])

    @staticmethod
    def assert_residue_matches_segment(res, seg):
        if isinstance(res, groups.Residue):
            res = [res]
        if isinstance(seg, groups.Segment):
            seg = itertools.cycle((seg,))
        for r, s in zip(res, seg):
            assert r.segindex == s.segindex
            assert r.segid == s.segid

    def test_move_residue(self, u):
        res = u.residues[0]
        source = u.segments[0]
        dest = u.segments[2]

        assert res in source.residues
        assert not res in dest.residues
        self.assert_residue_matches_segment(res, source)
        assert_equal(len(source.residues), 5)
        assert_equal(len(dest.residues), 5)

        res.segment = dest

        assert not res in source.residues
        assert res in dest.residues
        self.assert_residue_matches_segment(res, dest)
        assert_equal(len(source.residues), 4)
        assert_equal(len(dest.residues), 6)

    def test_move_residuegroup_single_segment(self, u):
        res = u.residues[[1, 3]]
        source = u.segments[0]
        dest = u.segments[2]

        for r in res:
            assert r in source.residues
            assert not r in dest.residues
        self.assert_residue_matches_segment(res, source)
        assert_equal(len(source.residues), 5)
        assert_equal(len(dest.residues), 5)

        res.segments = dest

        for r in res:
            assert not r in source.residues
            assert r in dest.residues
        self.assert_residue_matches_segment(res, dest)
        assert_equal(len(source.residues), 3)
        assert_equal(len(dest.residues), 7)

    def test_move_residuegroup_segmentgroup(self, u):
        res = u.residues[[1, 3]]
        source = u.segments[0]
        dest = u.segments[2] + u.segments[3]

        for r in res:
            assert r in source.residues
            assert not r in dest.residues
        self.assert_residue_matches_segment(res, source)
        assert_equal(len(source.residues), 5)
        assert_equal(len(dest[0].residues), 5)
        assert_equal(len(dest[1].residues), 5)

        res.segments = dest

        for r in res:
            assert not r in source.residues
            assert r in dest.residues
        self.assert_residue_matches_segment(res, dest)
        assert_equal(len(source.residues), 3)
        assert_equal(len(dest[0].residues), 6)
        assert_equal(len(dest[1].residues), 6)

    def test_move_residuegroup_segment_list(self, u):
        res = u.residues[[1, 3]]
        source = u.segments[0]
        dest = [u.segments[2], u.segments[3]]

        for r, d in zip(res, dest):
            assert r in source.residues
            assert not r in d.residues
        self.assert_residue_matches_segment(res, source)
        assert_equal(len(source.residues), 5)
        assert_equal(len(dest[0].residues), 5)
        assert_equal(len(dest[1].residues), 5)

        res.segments = dest

        for r, d in zip(res, dest):
            assert not r in source.residues
            assert r in d.residues
        self.assert_residue_matches_segment(res, dest)
        assert_equal(len(source.residues), 3)
        assert_equal(len(dest[0].residues), 6)
        assert_equal(len(dest[1].residues), 6)

    def test_move_residue_segmentgroup_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.residues[0], 'segment', u.segments[:4])

    def test_move_residue_list_TE(self, u):
        dest = [u.segments[3], u.segments[4]]
        with pytest.raises(TypeError):
            setattr(u.residues[0], 'segment', dest)

    def test_move_residuegroup_segmentgroup_VE(self, u):
        rg = u.residues[:3]
        sg = u.segments[1:]

        with pytest.raises(ValueError):
            setattr(rg, 'segments', sg)

    def test_move_residuegroup_list_VE(self, u):
        rg = u.residues[:2]
        sg = [u.segments[1], u.segments[2], u.segments[3]]

        with pytest.raises(ValueError):
            setattr(rg, 'segments', sg)

    def test_move_residue_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.residues[0], 'segment', 1)

    def test_move_residuegroup_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.residues[:3], 'segments', 4)

    def test_move_residuegroup_list_TE(self, u):
        with pytest.raises(TypeError):
            setattr(u.residues[:3], 'segments', [1, 2, 3])


class TestDownshiftArrays(object):
    @pytest.fixture()
    def square(self):
        return np.array([0, 1, 2, 0, 1, 2, 0, 1, 2])

    @pytest.fixture()
    def square_size(self):
        return 3

    @pytest.fixture()
    def square_result(self):
        return np.array([[0, 3, 6], [1, 4, 7], [2, 5, 8]])

    @pytest.fixture()
    def ragged(self):
        return np.array([0, 1, 2, 2, 0, 1, 2, 0, 1, 2])

    @pytest.fixture()
    def ragged_size(self):
        return 3

    @pytest.fixture()
    def ragged_result(self):
        return np.array([[0, 4, 7], [1, 5, 8], [2, 3, 6, 9]],
                        dtype=object)

    @staticmethod
    def assert_rows_match(a, b):
        for row_a, row_b in zip(a, b):
            assert_equal(row_a, row_b)

    # The array as a whole must be dtype object
    # While the subarrays must be integers
    def test_downshift_dtype_square(self, square, square_size):
        out = make_downshift_arrays(square, square_size)
        assert out.dtype == object
        assert out[0].dtype == np.intp

    def test_downshift_dtype_ragged(self, ragged, ragged_size):
        out = make_downshift_arrays(ragged, ragged_size)
        assert out.dtype == object
        assert out[0].dtype == np.intp

    # Check shape and size
    # Shape should be size N+1 as None is appended
    def test_shape_square(self, square, square_size):
        out = make_downshift_arrays(square, square_size)
        assert out.shape == (4,)
        assert out[-1] is None

    def test_shape_ragged(self, ragged, ragged_size):
        out = make_downshift_arrays(ragged, ragged_size)
        assert out.shape == (4,)
        assert out[-1] is None

    def test_contents_square(self, square, square_size, square_result):
        out = make_downshift_arrays(square, square_size)
        self.assert_rows_match(out, square_result)

    def test_contents_ragged(self, ragged, ragged_size, ragged_result):
        out = make_downshift_arrays(ragged, ragged_size)
        self.assert_rows_match(out, ragged_result)

    def test_missing_intra_values(self):
        out = make_downshift_arrays(
            np.array([0, 0, 2, 2, 3, 3]), 4)
        self.assert_rows_match(out,
                               np.array([np.array([0, 1]),
                                         np.array([], dtype=int),
                                         np.array([2, 3]),
                                         np.array([4, 5]),
                                         None], dtype=object))

    def test_missing_intra_values_2(self):
        out = make_downshift_arrays(
            np.array([0, 0, 3, 3, 4, 4]), 5)
        self.assert_rows_match(out,
                               np.array([np.array([0, 1]),
                                         np.array([], dtype=int),
                                         np.array([], dtype=int),
                                         np.array([2, 3]),
                                         np.array([4, 5]),
                                         None], dtype=object))

    def test_missing_end_values(self):
        out = make_downshift_arrays(np.array([0, 0, 1, 1, 2, 2]), 4)
        self.assert_rows_match(out,
                               np.array([np.array([0, 1]),
                                         np.array([2, 3]),
                                         np.array([4, 5]),
                                         np.array([], dtype=int),
                                         None], dtype=object))

    def test_missing_end_values_2(self):
        out = make_downshift_arrays(np.array([0, 0, 1, 1, 2, 2]), 6)
        self.assert_rows_match(out,
                               np.array([np.array([0, 1]),
                                         np.array([2, 3]),
                                         np.array([4, 5]),
                                         np.array([], dtype=int),
                                         np.array([], dtype=int),
                                         None], dtype=object))


class TestAddingResidues(object):
    """Tests for adding residues and segments to a Universe

    Adding Residues
    Adding Segments

    Adding Residue without an attr
    Adding Residue with ambiguous segment
    Adding Segment without an attr

    Adding Residue and moving atoms to it
    Adding Segment and moving residues to it
    """

    def test_add_segment_no_attrs(self):
        u = make_Universe()

        assert len(u.segments) == 5
        s = u.add_Segment()
        assert isinstance(s, MDAnalysis.core.groups.Segment)
        assert len(u.segments) == 6

    def test_add_residue_no_attrs(self):
        u = make_Universe()

        assert len(u.residues) == 25
        assert len(u.segments[1].residues) == 5

        res = u.add_Residue(segment=u.segments[1])

        assert len(u.residues) == 26
        assert len(u.segments[1].residues) == 6
        assert res in u.segments[1].residues

    def test_add_residue_no_attrs_one_segment(self):
        u = make_Universe(extras=[], size=(125, 25, 1))

        res = u.add_Residue()

        assert isinstance(res, MDAnalysis.core.groups.Residue)

    def test_add_Residue_ambiguous_segment_NDE(self):
        u = make_Universe()

        with pytest.raises(NoDataError):
            u.add_Residue()

    def test_add_Residue_missing_attr_NDE(self):
        u = make_Universe(('resids',))

        with pytest.raises(NoDataError):
            u.add_Residue(segment=u.segments[0])

    def test_add_Residue_NDE_message(self):
        # check error message asks for missing attr
        u = make_Universe(('resnames', 'resids'))

        try:
            u.add_Residue(segment=u.segments[0], resid=42)
        except NoDataError as e:
            assert 'resname' in str(e)
        else:
            raise AssertionError

    def test_add_Residue_NDE_message_2(self):
        # multiple missing attrs, check all get mentioned in error
        u = make_Universe(('resnames', 'resids'))

        try:
            u.add_Residue(segment=u.segments[0])
        except NoDataError as e:
            assert 'resname' in str(e)
            assert 'resid' in str(e)
        else:
            raise AssertionError

    def test_add_Residue_with_attrs(self):
        u = make_Universe(('resnames', 'resids'))

        r_new = u.add_Residue(segment=u.segments[0], resid=4321, resname='New')

        assert r_new.resid == 4321
        assert r_new.resname == 'New'

    def test_missing_attr_NDE_Segment(self):
        u = make_Universe(('segids',))
        with pytest.raises(NoDataError):
            u.add_Segment()

    def test_add_Segment_NDE_message(self):
        u = make_Universe(('segids',))

        try:
            u.add_Segment()
        except NoDataError as e:
            assert 'segid' in str(e)
        else:
            raise AssertionError

    def test_add_Segment_with_attr(self):
        u = make_Universe(('segids',))

        new_seg = u.add_Segment(segid='New')

        assert new_seg.segid == 'New'


class TestTopologyGuessed(object):
    @pytest.fixture()
    def names(self):
        return ta.Atomnames(np.array(['A', 'B', 'C'], dtype=object))

    @pytest.fixture()
    def types(self):
        return ta.Atomtypes(np.array(['X', 'Y', 'Z'], dtype=object),
                            guessed=True)

    @pytest.fixture()
    def resids(self):
        return ta.Resids(np.array([1]))

    @pytest.fixture()
    def resnames(self):
        return ta.Resnames(np.array(['ABC'], dtype=object),
                           guessed=True)

    @pytest.fixture()
    def top(self, names, types, resids, resnames):
        return Topology(n_atoms=3, n_res=1,
                        attrs=[names, types, resids, resnames])

    def test_guessed(self, names, types, resids, resnames, top):
        guessed = top.guessed_attributes

        assert types in guessed
        assert resnames in guessed
        assert not names in guessed
        assert not resids in guessed

    def test_read(self, names, types, resids, resnames, top):
        read = top.read_attributes

        assert names in read
        assert resids in read
        assert not types in read
        assert not resnames in read


class TestTopologyCreation(object):
    def test_make_topology_no_attrs(self):
        # should still make attrs list when attrs=None
        top = Topology()

        assert hasattr(top, 'attrs')
        assert isinstance(top.attrs, list)

    def test_resindex_VE(self):
        # wrong sized atom to residue array
        AR = np.arange(10)
        with pytest.raises(ValueError):
            Topology(n_atoms=5, atom_resindex=AR)

    def test_segindex_VE(self):
        # wrong sized residue to segment array
        AR = np.arange(5)
        RS = np.arange(10)
        with pytest.raises(ValueError):
            Topology(n_atoms=5, n_res=5, atom_resindex=AR,
                     residue_segindex=RS)



import numpy as np
from numpy.testing import assert_equal

from MDAnalysis.topology.base import squash_by, change_squash


class TestSquash(object):
    atom_resids = np.array([2, 2, 1, 1, 5, 5, 4, 4])
    atom_resnames = np.array(['A', 'A', 'B', 'B', 'C', 'C', 'D', 'D'],
                             dtype=object)

    def test_squash(self):
        atom_residx, resids, (resnames, ) = squash_by(self.atom_resids,
                                                      self.atom_resnames)

        assert_equal(atom_residx, np.array([1, 1, 0, 0, 3, 3, 2, 2]))
        assert_equal(resids, np.array([1, 2, 4, 5]))
        assert_equal(resnames, np.array(['B', 'A', 'D', 'C']))


class TestChangeSquash(object):
    def test_resid_squash(self):
        # System of 6 atoms in 3 residues
        # Residues 1 & 2 are Segid A, Residue 3 is Segid B
        # Resid 2 is repeated twice! Should be detected as 2 distinct residues
        resids = np.array([2, 2, 3, 3, 2, 2])
        resnames = np.array(['RsA', 'RsA', 'RsB', 'RsB', 'RsC', 'RsC'])
        segids = np.array(['A', 'A', 'A', 'A', 'B', 'B'])

        residx, (new_resids, new_resnames, new_segids) = change_squash(
            (resids, ), (resids, resnames, segids))

        assert_equal(residx, np.array([0, 0, 1, 1, 2, 2]))
        assert_equal(new_resids, np.array([2, 3, 2]))
        assert_equal(new_resnames, np.array(['RsA', 'RsB', 'RsC']))
        assert_equal(new_segids, np.array(['A', 'A', 'B']))

    def test_segid_squash(self):
        segids = np.array(['A', 'A', 'B'])

        segidx, (new_segids, ) = change_squash((segids, ), (segids, ))

        assert_equal(segidx, np.array([0, 0, 1]))
        assert_equal(new_segids, np.array(['A', 'B']))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from numpy.testing import assert_almost_equal

from MDAnalysisTests.datafiles import TXYZ, ARC, ARC_PBC

import MDAnalysis as mda


@pytest.fixture
def TXYZ_U():
    return mda.Universe(TXYZ)


@pytest.fixture
def ARC_U():
    return mda.Universe(ARC)


@pytest.fixture
def ARC_PBC_U():
    return mda.Universe(ARC_PBC)


def test_txyz_positions(TXYZ_U):
    assert_almost_equal(TXYZ_U.atoms.positions[0],
                        [-6.553398, -1.854369, 0.000000])


def test_arc_positions(ARC_U):
    assert_almost_equal(ARC_U.atoms.positions[0],
                        [-6.553398, -1.854369, 0.000000])


def test_arc_positions_frame_2(ARC_U):
    ARC_U.trajectory[1]

    assert_almost_equal(ARC_U.atoms.positions[0],
                        [-0.231579, -0.350841, -0.037475])


def test_arc_traj_length(ARC_U):
    assert len(ARC_U.trajectory) == 2


def test_arcpbc_traj_length(ARC_PBC_U):
    assert len(ARC_PBC_U.trajectory) == 3


def test_pbc_boxsize(ARC_PBC_U):
    ref_dimensions=[[ 38.860761,  38.860761,  38.860761,  90.000000,  90.000000,  90.000000],
                    [ 39.860761,  39.860761,  39.860761,  90.000000,  90.000000,  90.000000],
                    [ 40.860761,  40.860761,  40.860761,  90.000000,  90.000000,  90.000000]]

    for ref_box, ts in zip(ref_dimensions, ARC_PBC_U.trajectory):
        assert_almost_equal(ref_box, ts.dimensions, decimal=5)




# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

# Test the selection exporters in MDAnalysis.selections
# use StringIO and NamedStream to write to memory instead to temp files
import pytest
from io import StringIO

import re

import numpy as np
from numpy.testing import TestCase, assert_equal, assert_array_equal, dec

from MDAnalysis.tests.datafiles import PSF, DCD

import MDAnalysis
from MDAnalysis.lib.util import NamedStream


class _SelectionWriter(object):

    filename = None
    max_number = 357  # to keep fixtures smallish, only select CAs up to number 357

    @staticmethod
    @pytest.fixture()
    def universe():
        return MDAnalysis.Universe(PSF, DCD)

    @pytest.fixture()
    def namedfile(self):
        return NamedStream(StringIO(), self.filename)

    def _selection(self, universe):
        return universe.select_atoms("protein and name CA and bynum 1-{0}".format(self.max_number))

    def _write(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        g.write(namedfile, **kwargs)
        return g

    def _write_selection(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        g.write(namedfile, **kwargs)
        return g

    def _write_with(self, universe, namedfile, **kwargs):
        g = self._selection(universe)
        with self.writer(namedfile, **kwargs) as outfile:
            outfile.write(g)
        return g

    def test_write_bad_mode(self, universe, namedfile,):
        with pytest.raises(ValueError):
            self._write(universe, namedfile, name=self.ref_name, mode='a+')

    def test_write(self, universe, namedfile):
        self._write(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)

    def test_writeselection(self, universe, namedfile):
        self._write_selection(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)

    def test_write_with(self, universe, namedfile):
        self._write_with(universe, namedfile, name=self.ref_name)
        self._assert_selectionstring(namedfile)


def ndx2array(lines):
    """Convert Gromacs NDX text file lines to integer array"""
    return np.array(" ".join(lines).replace("\n", "").split(), dtype=int)


def lines2one(lines):
    """Join lines and squash all whitespace"""
    return " ".join(" ".join(lines).split())


class TestSelectionWriter_Gromacs(_SelectionWriter):

    writer = MDAnalysis.selections.gromacs.SelectionWriter
    filename = "CA.ndx"
    ref_name = "CA_selection"
    ref_indices = ndx2array(
        [ '5 22 46 65 84 103 122 129 141 153 160 170 \n',
          '177 199 206 220 237 247 264 284 303 320 335 357 \n',
          ]
        )

    def _assert_selectionstring(self, namedfile):
        header = namedfile.readline().strip()
        assert_equal(header, "[ {0} ]".format(self.ref_name),
                     err_msg="NDX file has wrong selection name")
        indices = ndx2array(namedfile.readlines())
        assert_array_equal(indices, self.ref_indices,
                           err_msg="indices were not written correctly")


class TestSelectionWriter_Charmm(_SelectionWriter):

    writer = MDAnalysis.selections.charmm.SelectionWriter
    filename = "CA.str"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one([
        """! MDAnalysis CHARMM selection
           DEFINE CA_selection SELECT -
           BYNUM 5 .or. BYNUM 22 .or. BYNUM 46 .or. BYNUM 65 .or. -
           BYNUM 84 .or. BYNUM 103 .or. BYNUM 122 .or. BYNUM 129 .or. -
           BYNUM 141 .or. BYNUM 153 .or. BYNUM 160 .or. BYNUM 170 .or. -
           BYNUM 177 .or. BYNUM 199 .or. BYNUM 206 .or. BYNUM 220 .or. -
           BYNUM 237 .or. BYNUM 247 .or. BYNUM 264 .or. BYNUM 284 .or. -
           BYNUM 303 .or. BYNUM 320 .or. BYNUM 335 .or. BYNUM 357 END
        """])

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(selectionstring, self.ref_selectionstring,
                     err_msg="Charmm selection was not written correctly")


class TestSelectionWriter_PyMOL(_SelectionWriter):

    writer = MDAnalysis.selections.pymol.SelectionWriter
    filename = "CA.pml"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one([
        """# MDAnalysis PyMol selection\n select CA_selection, \\
           index 5 | index 22 | index 46 | index 65 | index 84 | index 103 | \\
           index 122 | index 129 | index 141 | index 153 | index 160 | index 170 | \\
           index 177 | index 199 | index 206 | index 220 | index 237 | index 247 | \\
           index 264 | index 284 | index 303 | index 320 | index 335 | index 357
        """])

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(selectionstring, self.ref_selectionstring,
                     err_msg="PyMOL selection was not written correctly")


class TestSelectionWriter_VMD(_SelectionWriter):

    writer = MDAnalysis.selections.vmd.SelectionWriter
    filename = "CA.vmd"
    ref_name = "CA_selection"
    ref_selectionstring = lines2one([
        """# MDAnalysis VMD selection atomselect macro CA_selection {index 4 21 45 64 83 102 121 128 \\
           140 152 159 169 176 198 205 219 \\
           236 246 263 283 302 319 334 356 }
        """])

    def _assert_selectionstring(self, namedfile):
        selectionstring = lines2one(namedfile.readlines())
        assert_equal(selectionstring, self.ref_selectionstring,
                     err_msg="PyMOL selection was not written correctly")


def spt2array(line):
    """Get name of and convert Jmol SPT definition to integer array"""
    match = re.search(r'\@~(\w+) \(\{([\d\s]*)\}\)', line)
    return match.group(1), np.array(match.group(2).split(), dtype=int)


class TestSelectionWriter_Jmol(_SelectionWriter):

    writer = MDAnalysis.selections.jmol.SelectionWriter
    filename = "CA.spt"
    ref_name, ref_indices = spt2array(
        ( '@~ca ({4 21 45 64 83 102 121 128 140 152 159 169 176 198 205 219 236'
          ' 246 263 283 302 319 334 356});')
        )

    def _assert_selectionstring(self, namedfile):
        header, indices = spt2array(namedfile.readline())
        assert_equal(header, self.ref_name,
                     err_msg="SPT file has wrong selection name")
        assert_array_equal(indices, self.ref_indices,
                           err_msg="SPT indices were not written correctly")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import sys
import platform
import warnings
import MDAnalysis as mda
import pytest
import numpy as np
from numpy.testing import assert_equal
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    PRM,  # ache.prmtop
    PRM12,  # anti.top
    PRM7,  # tz2.truncoct.parm7.bz2
    PRMpbc,
    PRMNCRST,
    PRMNEGATIVE,
    PRMErr1,
    PRMErr2,
    PRMErr3,
    PRM_UreyBradley,
    PRM19SBOPC,
)


class TOPBase(ParserBase):
    parser = mda.topology.TOPParser.TOPParser
    expected_attrs = [
        "names", "types", "type_indices", "charges", "masses", "resnames",
        "bonds", "angles", "dihedrals", "impropers", "elements"
    ]
    expected_n_segments = 1

    def test_attr_size(self, top):
        assert len(top.names) == self.expected_n_atoms
        assert len(top.types) == self.expected_n_atoms
        assert len(top.type_indices) == self.expected_n_atoms
        assert len(top.charges) == self.expected_n_atoms
        assert len(top.masses) == self.expected_n_atoms
        assert len(top.resnames) == self.expected_n_residues
        assert len(top.bonds.values) == self.expected_n_bonds
        assert len(top.angles.values) == self.expected_n_angles
        assert len(top.dihedrals.values) == self.expected_n_dihedrals
        assert len(top.impropers.values) == self.expected_n_impropers

    def test_bonds_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].bonds) == self.expected_n_zero_bonds
        assert len(u.atoms[[self.atom_i]].bonds) == self.expected_n_i_bonds

    def test_angles_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].angles) == self.expected_n_zero_angles
        assert len(u.atoms[[self.atom_i]].angles) == self.expected_n_i_angles

    def test_dihedrals_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].dihedrals) == self.expected_n_zero_dihedrals
        assert len(u.atoms[[self.atom_i]].dihedrals) == \
            self.expected_n_i_dihedrals

    def test_impropers_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].impropers) == self.expected_n_zero_impropers
        assert len(u.atoms[[self.atom_i]].impropers) == \
            self.expected_n_i_impropers

    def test_bonds_identity(self, top):
        vals = top.bonds.values
        for bond in self.atom_zero_bond_values:
            assert (bond in vals) or (bond[::-1] in vals)
        for bond in self.atom_i_bond_values:
            assert (bond in vals) or (bond[::-1] in vals)

    def test_angles_identity(self, top):
        vals = top.angles.values
        for ang in self.atom_zero_angle_values:
            assert (ang in vals) or (ang[::-1] in vals)
        for ang in self.atom_i_angle_values:
            assert (ang in vals) or (ang[::-1] in vals)

    def test_dihedrals_identity(self, top):
        vals = top.dihedrals.values
        for dih in self.atom_zero_dihedral_values:
            assert (dih in vals) or (dih[::-1] in vals)
        for dih in self.atom_i_dihedral_values:
            assert (dih in vals) or (dih[::-1] in vals)

    def test_impropers_identity(self, top):
        vals = top.impropers.values
        for imp in self.atom_zero_improper_values:
            assert (imp in vals) or (imp[::-1] in vals)
        for imp in self.atom_i_improper_values:
            assert (imp in vals) or (imp[::-1] in vals)

    def test_angle_atoms_bonded(self, top):
        vals = top.bonds.values
        for ang in top.angles.values:
            for b in ((ang[0], ang[1]), (ang[1], ang[2])):
                assert (b in vals) or (b[::-1] in vals)

    def test_dihedral_atoms_bonded(self, top):
        vals = top.bonds.values
        for dih in top.dihedrals.values:
            for b in ((dih[0], dih[1]), (dih[1], dih[2]), (dih[2], dih[3])):
                assert (b in vals) or (b[::-1] in vals)

    def test_improper_atoms_bonded(self, top):
        vals = top.bonds.values
        for imp in top.impropers.values:
            forward = ((imp[0], imp[2]), (imp[1], imp[2]), (imp[2], imp[3]))
            backward = ((imp[0], imp[1]), (imp[1], imp[2]), (imp[1], imp[3]))
            for a, b in zip(forward, backward):
                assert ((b in vals) or (b[::-1] in vals) or
                        (a in vals) or (a[::-1] in vals))

    def test_elements(self, top):
        """Tests elements attribute.

        If elements present, loops over ranges of the topology elements list
        and compared against a provided list of expected values.
        Otherwise, checks that elements are not in the topology attributes.
        """

        if self.expected_elems:
            for erange, evals in zip(self.elems_ranges, self.expected_elems):
                assert_equal(top.elements.values[erange[0]:erange[1]], evals,
                             "unexpected element match")
        else:
            assert not hasattr(top, 'elements'), 'Unexpected elements attr'


class TestPRMParser(TOPBase):
    ref_filename = PRM
    # Does not contain an ATOMIC_NUMBER record, so no elements
    expected_attrs = [
        "names", "types", "type_indices", "charges", "masses", "resnames",
        "bonds", "angles", "dihedrals", "impropers"
    ]
    expected_n_atoms = 252
    expected_n_residues = 14
    expected_n_bonds = 259
    expected_n_angles = 456
    expected_n_dihedrals = 673
    expected_n_impropers = 66
    atom_i = 79
    expected_n_zero_bonds = 4
    expected_n_i_bonds = 3
    expected_n_zero_angles = 9
    expected_n_i_angles = 9
    expected_n_zero_dihedrals = 14
    expected_n_i_dihedrals = 15
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 4
    atom_zero_bond_values = ((0, 4), (0, 1), (0, 2), (0, 3))
    atom_i_bond_values = ((79, 80), (79, 83), (77, 79))
    atom_zero_angle_values = ((0, 4, 6), (0, 4, 10), (3, 0, 4),
                              (2, 0, 3), (2, 0, 4), (1, 0, 2),
                              (1, 0, 3), (1, 0, 4), (0, 4, 5))
    atom_i_angle_values = ((80, 79, 83), (77, 79, 80), (77, 79, 83),
                           (74, 77, 79), (79, 80, 81), (79, 80, 82),
                           (79, 83, 84), (79, 83, 85), (78, 77, 79))
    atom_zero_dihedral_values = ((0, 4, 10, 11), (0, 4, 10, 12),
                                 (3, 0, 4, 5), (3, 0, 4, 6),
                                 (3, 0, 4, 10), (2, 0, 4, 5),
                                 (2, 0, 4, 6), (2, 0, 4, 10),
                                 (1, 0, 4, 5), (1, 0, 4, 6),
                                 (1, 0, 4, 10), (0, 4, 6, 7),
                                 (0, 4, 6, 8), (0, 4, 6, 9))
    atom_i_dihedral_values = ((71, 74, 77, 79), (74, 77, 79, 80),
                              (74, 77, 79, 83), (75, 74, 77, 79),
                              (76, 74, 77, 79), (77, 79, 80, 81),
                              (77, 79, 80, 82), (77, 79, 83, 84),
                              (77, 79, 83, 85), (78, 77, 79, 80),
                              (78, 77, 79, 83), (80, 79, 83, 84),
                              (80, 79, 83, 85), (81, 80, 79, 83),
                              (82, 80, 79, 83))
    atom_zero_improper_values = ()
    atom_i_improper_values = ((74, 79, 77, 78), (77, 80, 79, 83),
                              (79, 81, 80, 82), (79, 84, 83, 85))
    expected_elems = None


class TestPRM12Parser(TOPBase):
    ref_filename = PRM12
    expected_n_atoms = 8923
    expected_n_residues = 2861
    expected_n_bonds = 8947
    expected_n_angles = 756
    expected_n_dihedrals = 1128
    expected_n_impropers = 72
    expected_n_zero_bonds = 1
    expected_n_i_bonds = 4
    expected_n_zero_angles = 1
    expected_n_i_angles = 12
    expected_n_zero_dihedrals = 3
    expected_n_i_dihedrals = 28
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 1
    atom_i = 335
    ref_proteinatoms = 0
    atom_zero_bond_values = ((0, 1),)
    atom_i_bond_values = ((335, 337), (335, 354),
                          (334, 335), (335, 336))
    atom_zero_angle_values = ((0, 1, 2),)
    atom_i_angle_values = ((337, 335, 354), (335, 337, 338),
                           (335, 337, 351), (335, 354, 352),
                           (334, 335, 337), (334, 335, 354),
                           (332, 334, 335), (336, 335, 337),
                           (336, 335, 354), (335, 354, 355),
                           (335, 354, 356), (334, 335, 336))
    atom_zero_dihedral_values = ((0, 1, 2, 3), (0, 1, 2, 4),
                                 (0, 1, 2, 5))
    atom_i_dihedral_values = ((329, 332, 334, 335), (332, 334, 335, 336),
                              (332, 334, 335, 337), (332, 334, 335, 354),
                              (332, 352, 354, 335), (333, 332, 334, 335),
                              (334, 335, 337, 338), (334, 335, 337, 351),
                              (334, 335, 354, 352), (334, 335, 354, 355),
                              (334, 335, 354, 356), (335, 334, 332, 352),
                              (335, 337, 338, 339), (335, 337, 338, 340),
                              (335, 337, 351, 341), (335, 337, 351, 350),
                              (335, 354, 352, 353), (335, 354, 352, 357),
                              (336, 335, 337, 338), (336, 335, 337, 351),
                              (336, 335, 354, 352), (336, 335, 354, 355),
                              (336, 335, 354, 356), (337, 335, 354, 352),
                              (337, 335, 354, 355), (337, 335, 354, 356),
                              (338, 337, 335, 354), (351, 337, 335, 354))
    atom_zero_improper_values = ()
    atom_i_improper_values = ((335, 337, 338, 351),)
    elems_ranges = [[0, 36], [351, 403]]
    expected_elems = [np.array(["H", "O", "C", "H", "H", "C", "H", "O", "C",
                                "H", "N", "C", "H", "N", "C", "C", "O", "N",
                                "H", "C", "N", "H", "H", "N", "C", "C", "H",
                                "C", "H", "H", "O", "P", "O", "O", "O", "C"],
                      dtype=object),
                      np.array(["C", "C", "H", "C", "H", "H", "O", "P", "O",
                                "O", "O", "C", "H", "H", "C", "H", "O", "C",
                                "H", "N", "C", "H", "N", "C", "C", "O", "N",
                                "H", "C", "N", "H", "H", "N", "C", "C", "H",
                                "C", "H", "H", "O", "H", "Na", "Na", "Na",
                                "Na", "Na", "Na", "Na", "Na", "O", "H", "H"],
                      dtype=object)]


class TestParm7Parser(TOPBase):
    ref_filename = PRM7
    # Does not contain an ATOMIC_NUMBER record, so no elements
    expected_attrs = [
        "names", "types", "type_indices", "charges", "masses", "resnames",
        "bonds", "angles", "dihedrals", "impropers"
    ]
    expected_n_atoms = 5827
    expected_n_residues = 1882
    expected_n_bonds = 5834
    expected_n_angles = 402
    expected_n_dihedrals = 602
    expected_n_impropers = 55
    atom_i = 135
    expected_n_zero_bonds = 4
    expected_n_i_bonds = 4
    expected_n_zero_angles = 9
    expected_n_i_angles = 13
    expected_n_zero_dihedrals = 14
    expected_n_i_dihedrals = 27
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 2
    atom_zero_bond_values = ((0, 4), (0, 1), (0, 2), (0, 3))
    atom_i_bond_values = ((135, 137), (135, 155), (133, 135),
                          (135, 136))
    atom_zero_angle_values = ((0, 4, 6), (0, 4, 11), (3, 0, 4),
                              (2, 0, 3), (2, 0, 4), (1, 0, 2),
                              (1, 0, 3), (1, 0, 4), (0, 4, 5))
    atom_i_angle_values = ((131, 133, 135), (137, 135, 155),
                           (135, 137, 140), (135, 155, 156),
                           (135, 155, 157), (133, 135, 137),
                           (133, 135, 155), (136, 135, 137),
                           (136, 135, 155), (135, 137, 138),
                           (135, 137, 139), (134, 133, 135),
                           (133, 135, 136))
    atom_zero_dihedral_values = ((0, 4, 6, 7), (0, 4, 6, 8),
                                 (0, 4, 6, 9), (0, 4, 11, 12),
                                 (0, 4, 11, 13), (1, 0, 4, 5),
                                 (1, 0, 4, 6), (1, 0, 4, 11),
                                 (2, 0, 4, 5), (2, 0, 4, 6),
                                 (2, 0, 4, 11), (3, 0, 4, 5),
                                 (3, 0, 4, 6), (3, 0, 4, 11))
    atom_i_dihedral_values = ((113, 131, 133, 135), (131, 133, 135, 136),
                              (131, 133, 135, 137), (131, 133, 135, 155),
                              (132, 131, 133, 135), (133, 135, 137, 138),
                              (133, 135, 137, 139), (133, 135, 137, 140),
                              (133, 135, 155, 156), (133, 135, 155, 157),
                              (134, 133, 135, 136), (134, 133, 135, 137),
                              (134, 133, 135, 155), (135, 137, 140, 141),
                              (135, 137, 140, 154), (135, 155, 157, 158),
                              (135, 155, 157, 159), (136, 135, 137, 138),
                              (136, 135, 137, 139), (136, 135, 137, 140),
                              (136, 135, 155, 156), (136, 135, 155, 157),
                              (137, 135, 155, 156), (137, 135, 155, 157),
                              (138, 137, 135, 155), (139, 137, 135, 155),
                              (140, 137, 135, 155))
    atom_zero_improper_values = ()
    atom_i_improper_values = ((131, 135, 133, 134), (135, 157, 155, 156))
    expected_elems = None


class TestPRM2(TOPBase):
    ref_filename = PRMpbc
    # Does not contain an ATOMIC_NUMBER record, so no elements
    expected_attrs = [
        "names", "types", "type_indices", "charges", "masses", "resnames",
        "bonds", "angles", "dihedrals", "impropers"
    ]
    expected_n_atoms = 5071
    expected_n_residues = 1686
    ref_proteinatoms = 22
    expected_n_bonds = 5070
    expected_n_angles = 36
    expected_n_dihedrals = 41
    expected_n_impropers = 4
    atom_i = 14
    expected_n_zero_bonds = 1
    expected_n_i_bonds = 3
    expected_n_zero_angles = 3
    expected_n_i_angles = 8
    expected_n_zero_dihedrals = 2
    expected_n_i_dihedrals = 18
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 2
    atom_zero_bond_values = ((0, 1),)
    atom_i_bond_values = ((14, 15), (14, 16), (8, 14))
    atom_zero_angle_values = ((0, 1, 2), (0, 1, 3), (0, 1, 4))
    atom_i_angle_values = ((15, 14, 16), (14, 16, 18), (10,  8, 14),
                           (8, 14, 15), (8, 14, 16), (6,  8, 14),
                           (14, 16, 17), (9,  8, 14))
    atom_zero_dihedral_values = ((0, 1, 4, 5), (0, 1, 4, 6))
    atom_i_dihedral_values = ((4, 6, 8, 14), (6, 8, 14, 15),
                              (6, 8, 14, 16), (7, 6, 8, 14),
                              (8, 14, 16, 17), (8, 14, 16, 18),
                              (9, 8, 14, 15), (9, 8, 14, 16),
                              (10, 8, 14, 15), (10, 8, 14, 16),
                              (11, 10, 8, 14), (12, 10, 8, 14),
                              (13, 10, 8, 14), (14, 16, 18, 19),
                              (14, 16, 18, 20), (14, 16, 18, 21),
                              (15, 14, 16, 17), (15, 14, 16, 18))
    atom_zero_improper_values = ()
    atom_i_improper_values = ((8, 16, 14, 15), (14, 18, 16, 17))
    expected_elems = None


class TestPRMNCRST(TOPBase):
    # Test case of PARM7 with no non-hydrogen including dihedrals
    ref_filename = PRMNCRST
    expected_n_atoms = 6
    expected_n_residues = 1
    ref_proteinatoms = 6
    expected_n_bonds = 5
    expected_n_angles = 7
    expected_n_dihedrals = 3
    expected_n_impropers = 0
    atom_i = 4
    expected_n_zero_bonds = 1
    expected_n_i_bonds = 2
    expected_n_zero_angles = 3
    expected_n_i_angles = 4
    expected_n_zero_dihedrals = 1
    expected_n_i_dihedrals = 3
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 0
    atom_zero_bond_values = ((0, 1),)
    atom_i_bond_values = ((1, 4), (4, 5))
    atom_zero_angle_values = ((0, 1, 2), (0, 1, 3), (0, 1, 4))
    atom_i_angle_values = ((0, 1, 4), (1, 4, 5), (2, 1, 4), (3, 1, 4))
    atom_zero_dihedral_values = ((0, 1, 4, 5),)
    atom_i_dihedral_values = ((0, 1, 4, 5), (2, 1, 4, 5), (3, 1, 4, 5))
    atom_zero_improper_values = ()
    atom_i_improper_values = ()
    elems_ranges = [[0, 6], ]
    expected_elems = [np.array(["H", "C", "H", "H", "C", "O"], dtype=object), ]


class TestPRMNCRST_negative(TOPBase):
    # Same as above but with negative ATOMIC_NUMBER values (Issue 2306)
    ref_filename = PRMNEGATIVE
    expected_n_atoms = 6
    expected_n_residues = 1
    ref_proteinatoms = 6
    expected_n_bonds = 5
    expected_n_angles = 7
    expected_n_dihedrals = 3
    expected_n_impropers = 0
    atom_i = 4
    expected_n_zero_bonds = 1
    expected_n_i_bonds = 2
    expected_n_zero_angles = 3
    expected_n_i_angles = 4
    expected_n_zero_dihedrals = 1
    expected_n_i_dihedrals = 3
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 0
    atom_zero_bond_values = ((0, 1),)
    atom_i_bond_values = ((1, 4), (4, 5))
    atom_zero_angle_values = ((0, 1, 2), (0, 1, 3), (0, 1, 4))
    atom_i_angle_values = ((0, 1, 4), (1, 4, 5), (2, 1, 4), (3, 1, 4))
    atom_zero_dihedral_values = ((0, 1, 4, 5),)
    atom_i_dihedral_values = ((0, 1, 4, 5), (2, 1, 4, 5), (3, 1, 4, 5))
    atom_zero_improper_values = ()
    atom_i_improper_values = ()
    elems_ranges = [[0, 6], ]
    expected_elems = [np.array(["H", "", "H", "H", "C", ""], dtype=object), ]


class TestPRMEP(TOPBase):
    # Issue #2449
    ref_filename = PRM19SBOPC
    expected_n_atoms = 46
    expected_n_residues = 9
    ref_proteinatoms = 23
    expected_n_bonds = 45
    expected_n_angles = 36
    expected_n_dihedrals = 41
    expected_n_impropers = 4
    atom_i = 25  # testing EPW atom
    expected_n_zero_bonds = 1
    expected_n_i_bonds = 1
    expected_n_zero_angles = 3
    expected_n_i_angles = 0
    expected_n_zero_dihedrals = 2
    expected_n_i_dihedrals = 0
    expected_n_zero_impropers = 0
    expected_n_i_impropers = 0
    atom_zero_bond_values = ((0, 1),)
    atom_i_bond_values = ((25, 22),)
    atom_zero_angle_values = ((0, 1, 2), (0, 1, 3), (0, 1, 4))
    atom_i_angle_values = ()
    atom_zero_dihedral_values = ((0, 1, 4, 5), (0, 1, 4, 6))
    atom_i_dihedral_values = ()
    atom_zero_improper_values = ()
    atom_i_improper_values = ()
    elems_ranges = [[0, 8], [20, 28]]
    expected_elems = [np.array(["H", "C", "H", "H", "C", "O", "N", "H"],
                      dtype=object),
                      np.array(["H", "H", "O", "H", "H", "", "O", "H"],
                      dtype=object)]


class TestErrorsAndWarnings(object):

    ATOMIC_NUMBER_MSG = ("ATOMIC_NUMBER record not found, elements attribute "
                         "will not be populated")
    MISSING_ELEM_MSG = ("Unknown ATOMIC_NUMBER value found for some atoms, "
                        "these have been given an empty element record")
    COORDINATE_READER_MSG = ("No coordinate reader found")

    @pytest.mark.parametrize("parm,errmatch", (
        [PRMErr1, "%VE Missing in header"],
        [PRMErr2, "'TITLE' missing in header"],
        [PRM_UreyBradley, "Chamber-style TOP file"]
    ))
    def test_value_errors(self, parm, errmatch):
        with pytest.raises(ValueError, match=errmatch):
            u = mda.Universe(parm)

    def test_flag_index_error(self):
        with pytest.raises(IndexError, match="%FLAG section not found"):
            u = mda.Universe(PRMErr3)

    @pytest.mark.parametrize("parm, errmsgs", (
        [PRM, [ATOMIC_NUMBER_MSG, COORDINATE_READER_MSG]],
        [PRM7, [ATOMIC_NUMBER_MSG, COORDINATE_READER_MSG]],
        [PRMpbc, [ATOMIC_NUMBER_MSG, COORDINATE_READER_MSG]],
        [PRMNEGATIVE, [MISSING_ELEM_MSG, COORDINATE_READER_MSG]],
        [PRM19SBOPC, [MISSING_ELEM_MSG, COORDINATE_READER_MSG]]
    ))
    def test_warning(self, parm, errmsgs):
        with pytest.warns(UserWarning) as record:
            u = mda.Universe(parm)

        messages = [rec.message.args[0] for rec in record]

        for msg in errmsgs:
            assert any(msg in recmsg for recmsg in messages)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from numpy.testing import assert_almost_equal

from MDAnalysisTests.datafiles import GSD

import MDAnalysis as mda
import os


@pytest.fixture
def GSD_U():
    return mda.Universe(GSD)

def test_gsd_positions(GSD_U):
    # first frame first particle
    ts = GSD_U.trajectory[0]
    assert_almost_equal(GSD_U.atoms.positions[0],
                        [ -5.4000001 , -10.19999981, -10.19999981])
    # second frame first particle
    ts = GSD_U.trajectory[1]
    assert_almost_equal(GSD_U.atoms.positions[0],
                        [ -5.58348083,  -9.98546982, -10.17657185])

def test_gsd_n_frames(GSD_U):
    assert len(GSD_U.trajectory) == 2

def test_gsd_dimensions(GSD_U):
    ts = GSD_U.trajectory[0]
    assert_almost_equal(ts.dimensions,
                        [ 21.60000038,21.60000038,21.60000038,90.,90.,90.])

def test_gsd_data_step(GSD_U):
    assert GSD_U.trajectory[0].data['step'] == 0
    assert GSD_U.trajectory[1].data['step'] == 500



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import re

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_almost_equal, assert_allclose

import MDAnalysis as mda
from MDAnalysis.analysis import helix_analysis as hel
from MDAnalysisTests.datafiles import (GRO, XTC, PSF, DCD, PDB_small,
                                       HELANAL_BENDING_MATRIX,
                                       HELANAL_BENDING_MATRIX_SUBSET,
                                       XYZ)

# reference data from old helix analysis of a single PDB file:
#   data = MDAnalysis.analysis.helanal.helanal_main(PDB_small,
#                    select="name CA and resnum 161-187")
# keys are renamed and local screw angles now use a different
# algorithm
HELANAL_SINGLE_DATA = {
    'global_tilts': np.rad2deg(1.3309656332019535),
    'local_heights summary': np.array([1.5286051,  0.19648294,  0.11384312],
                                      dtype=np.float32),
    'local_bends':
        np.array([3.44526005,   4.85425806,   4.69548464,   2.39473653,
                  3.56172442,   3.97128344,   3.41563916,   1.86140978,
                  5.22997046,   5.41381264,  27.49601364,  39.69839478,
                  35.05921936,  21.78928566,   9.8632431,   8.80066967,
                  5.5344553,   6.14356709,  10.15450764,  11.07686138,
                  9.23541832], dtype=np.float32),
    'local_nres_per_turn summary': np.array([3.64864163, 0.152694,
                                             0.1131402]),
    'local_twists summary': np.array([98.83011627, 4.08171701, 3.07253003],
                                     dtype=np.float32),
    'local_twists':
        np.array([97.23709869,   99.09676361,   97.25350952,  101.76019287,
                  100.42689514,   97.08784485,   97.07430267,   98.33553314,
                  97.86578369,   95.45792389,   97.10089111,   95.26415253,
                  87.93136597,  108.38458252,   95.27167511,  104.01931763,
                  100.7199707,  101.48034668,   99.64170074,   94.78946686,
                  102.50147247,   97.25154877,  104.54204559,  101.42829895],
                 dtype=np.float32),
}


def read_bending_matrix(fn):
    """Read helanal_bending_matrix.dat into dict of numpy arrays.

    This is a quick and dirty parser with no error checking.

    Format::

      Mean
         0.0    5.7   10.9 ....
         .

      SD
        ....
        .

      ABDEV
       ....
       .

    Returns
    -------
       data : dict
           dictionary with keys 'Mean', 'SD', 'ABDEV' and NxN matrices.
    """
    data = {}
    current = None
    with open(fn) as bendingmatrix:
        for line in bendingmatrix:
            line = line.strip()
            if not line:
                continue
            if re.match(r"\D+", line):
                label = line.split()[0]
                current = data[label] = []
            else:
                current.append([float(x) for x in line.split()])
    for k, v in data.items():
        data[k] = np.array(v)
    return data


def test_local_screw_angles_plane_circle():
    """
    Test the trivial case of a circle in the xy-plane.
    The global axis is the x-axis and ref axis is the y-axis,
    so angles should be calculated to the xy-plane.
    """
    angdeg = np.arange(0, 360, 12, dtype=np.int32)
    angrad = np.deg2rad(angdeg, dtype=np.float64)
    xyz = np.array([[np.cos(a), np.sin(a), 0] for a in angrad],
                   dtype=np.float64)
    xyz[15, 1] = 0  # because np.sin(np.deg2rad(180)) = 1e-16 ?!
    screw = hel.local_screw_angles([1, 0, 0], [0, 1, 0], xyz)
    correct = np.zeros_like(angdeg)
    correct[(angdeg > 180)] = 180
    assert_almost_equal(screw, correct)


def test_local_screw_angles_ortho_circle():
    """
    Test the trivial case of a circle in the xy-plane.
    The global axis is the x-axis and ref axis is the z-axis,
    so angles should be calculated to the xz-plane.
    """
    angdeg = np.arange(0, 360, 12, dtype=np.int32)
    angrad = np.deg2rad(angdeg, dtype=np.float64)
    xyz = np.array([[np.cos(a), np.sin(a), 0] for a in angrad],
                   dtype=np.float64)
    xyz[15, 1] = 0  # because np.sin(np.deg2rad(180)) = 1e-16 ?!
    screw = hel.local_screw_angles([1, 0, 0], [0, 0, 1], xyz)
    correct = np.zeros_like(angdeg)
    correct[(angdeg < 180)] = 90
    correct[(angdeg > 180)] = -90
    correct[0] = correct[15] = 0
    assert_almost_equal(screw, correct)


def test_local_screw_angles_around_circle():
    """
    Test an orthogonal circle in the xz-plane.
    The global axis is the y-axis and ref axis is the x-axis,
    so angles should be calculated to the xy-plane.
    """
    # circle in xz-plane
    angdeg = np.arange(0, 360, 12, dtype=int)
    angrad = np.deg2rad(angdeg)
    xyz = np.array([[np.cos(a), 0, np.sin(a)] for a in angrad])
    screw = hel.local_screw_angles([0, 1, 0], [1, 0, 0], xyz)
    angdeg[-14:] = -angdeg[1:15][::-1]
    angdeg[-15] = 180
    assert_almost_equal(screw, angdeg)


def test_local_screw_angles_around_circle_rising():
    """
    Test a circle in the xz-plane, rising on the y-axis.
    The global axis is the y-axis and ref axis is the x-axis,
    so angles should be calculated to the xy-plane.
    The y-axis contribution should be removed so angles should
    be to the circle.
    """
    # circle in xz-plane
    angdeg = np.arange(0, 360, 12, dtype=int)
    angrad = np.deg2rad(angdeg)
    xyz = np.array([[np.cos(a), i, np.sin(a)] for i, a in enumerate(angrad)])
    screw = hel.local_screw_angles([0, 1, 0], [1, 0, 0], xyz)
    angdeg[-14:] = -angdeg[1:15][::-1]
    angdeg[-15] = 180
    assert_almost_equal(screw, angdeg)


def test_local_screw_angles_parallel_axes():
    """
    Test that if global_axis and ref_axis are parallel, it
    picks another one instead. global_axis is the x-axis,
    so the eventual replacement ref_axis should be the z-axis,
    so angles should be calculated to the xz-plane.
    """
    xyz = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    angles = hel.local_screw_angles([1, 0, 0], [-1, 0, 0], xyz)
    assert_almost_equal(angles, [0, 90, 0])


@pytest.fixture()
def zigzag():
    r"""
    x-axis is from bottom to up in plane of screen
    z-axis is left to right ->
         x    x    x    x    x
        / \  / \  / \  / \  /
      x    x    x    x    x
    """
    n_atoms = 100
    u = mda.Universe.empty(100, atom_resindex=np.arange(n_atoms),
                           trajectory=True)
    xyz = np.array(list(zip([1, -1]*(n_atoms//2),  # x \in {0, 1}
                            [0]*n_atoms,  # y == 0
                            range(n_atoms))))  # z rises continuously
    u.load_new(xyz)
    return u


@pytest.mark.parametrize('ref_axis,screw_angles', [
    # input vectors zigzag between [-1, 0, 0] and [1, 0, 0]
    # global axis is z-axis
    ([0, 0, 1], [180, 0]),  # calculated to x-z plane
    ([1, 0, 0], [180, 0]),  # calculated to x-z plane
    ([-1, 0, 0], [0, 180]),  # calculated to x-z plane upside down
    ([0, 1, 0], [90, -90]),  # calculated to y-z plane
    ([0, -1, 0], [-90, 90]),  # calculated to x-z plane upside down
    # calculated to diagonal xy-z plane rotating around
    ([1, 1, 0], [135, -45]),
    ([-1, 1, 0], [45, -135]),
    ([1, -1, 0], [-135, 45]),
    ([-1, -1, 0], [-45, 135]),
    # calculated to diagonal xyz-z plane w/o z contribution
    ([1, 1, 1], [135, -45]),
    ([1, 1, -1], [135, -45]),
    ([1, -1, 1], [-135, 45]),
    ([-1, 1, 1], [45, -135]),
    ([-1, -1, 1], [-45, 135]),
    ([-1, -1, -1], [-45, 135]),
])
def test_helix_analysis_zigzag(zigzag, ref_axis, screw_angles):
    properties = hel.helix_analysis(zigzag.atoms.positions,
                                    ref_axis=ref_axis)
    assert_almost_equal(properties['local_twists'], 180, decimal=4)
    assert_almost_equal(properties['local_nres_per_turn'], 2, decimal=4)
    assert_almost_equal(properties['global_axis'],
                        [0, 0, -1], decimal=4)
    # all 0 vectors
    assert_almost_equal(properties['local_axes'], 0, decimal=4)
    assert_almost_equal(properties['local_bends'], 0, decimal=4)
    assert_almost_equal(properties['all_bends'], 0, decimal=4)
    assert_almost_equal(properties['local_heights'], 0, decimal=4)
    assert_almost_equal(properties['local_helix_directions'][0::2],
                        [[-1, 0, 0]]*49, decimal=4)
    assert_almost_equal(properties['local_helix_directions'][1::2],
                        [[1, 0, 0]]*49, decimal=4)
    origins = zigzag.atoms.positions[1:-1].copy()
    origins[:, 0] = 0
    assert_almost_equal(properties['local_origins'], origins, decimal=4)
    assert_almost_equal(properties['local_screw_angles'],
                        screw_angles*49, decimal=4)


def square_oct(n_rep=10):
    """
    square-octagon-square-octagon

    y- and z- coordinates create the square and octogon.
    x-coordinates increase continuously.
    """
    # square-octagon-square-octagon
    sq2 = 0.5 ** 0.5
    square = [(1, 0), (0, 1), (-1, 0), (0, -1)]
    octagon = [(1, 0), (sq2, sq2), (0, 1), (-sq2, sq2),
               (-1, 0), (-sq2, -sq2), (0, -1), (sq2, -sq2)]
    shapes = (square+octagon)*n_rep
    xyz = np.array(list(zip(np.arange(len(shapes)),
                            *zip(*shapes))))
    n_atoms = len(xyz)
    u = mda.Universe.empty(n_atoms, trajectory=True)
    u.load_new(xyz)
    return u


def test_helix_analysis_square_oct():
    n_rep = 10
    u = square_oct(n_rep=n_rep)
    n_atoms = len(u.atoms)

    properties = hel.helix_analysis(u.atoms.positions,
                                    ref_axis=[0, 0, 1])
    twist_trans = [102.76438, 32.235607]
    twists = ([90]*2 + twist_trans + [45]*6 + twist_trans[::-1]) * n_rep
    assert_almost_equal(properties['local_twists'], twists[:n_atoms-3],
                        decimal=4)
    res_trans = [3.503159, 11.167775]
    res = ([4]*2 + res_trans + [8]*6 + res_trans[::-1]) * n_rep
    assert_almost_equal(properties['local_nres_per_turn'], res[:n_atoms-3],
                        decimal=4)
    assert_almost_equal(properties['global_axis'],
                        [-1, 0, 0], decimal=3)
    assert_almost_equal(properties['local_axes']-[1, 0, 0], 0, decimal=4)
    assert_almost_equal(properties['local_bends'], 0, decimal=4)
    assert_almost_equal(properties['all_bends'], 0, decimal=4)
    assert_almost_equal(properties['local_heights'], 1, decimal=4)

    loc_rot = [[0.,  0.,  1.],
               [0., -1.,  0.],
               [0.,  0., -1.],
               [0.,  0.97528684,  0.2209424],  # the transition square->oct
               [0.,  0.70710677,  0.70710677],  # 0.5 ** 0.5
               [0.,  0.,  1.],
               [0., -0.70710677,  0.70710677],
               [0., -1.,  0.],
               [0., -0.70710677, -0.70710677],
               [0.,  0., -1.],
               [0.,  0.70710677, -0.70710677],
               [0.,  0.97528684, -0.2209424]] * n_rep
    assert_almost_equal(properties['local_helix_directions'],
                        loc_rot[:n_atoms-2], decimal=4)

    origins = u.atoms.positions.copy()[1:-1]
    origins[:, 1:] = ([[0.,   0.],  # square
                       [0.,   0.],
                       [0.,  -0.33318555],  # square -> octagon
                       [-1.7878988,  -0.6315732],  # square -> octagon
                       [0.,   0.],  # octagon
                       [0.,   0.],
                       [0.,   0.],
                       [0.,   0.],
                       [0.,   0.],
                       [0.,   0.],
                       [-1.3141878,   1.3141878],  # octagon -> square
                       [0.34966463,   0.14732757]]*n_rep)[:len(origins)]
    assert_almost_equal(properties['local_origins'], origins,
                        decimal=4)

    # calculated to the x-y plane
    # all input vectors (loc_rot) are in y-z plane
    screw = [0, 90, 180,  # square
             -77.236,  # transition
             -45, 0, 45, 90, 135, 180, -135,  # octagon
             -102.764]*n_rep

    # not quite 0, comes out as 1.32e-06
    assert_almost_equal(properties['local_screw_angles'], screw[:-2],
                        decimal=3)


class TestHELANAL(object):

    @pytest.fixture()
    def psf_ca(self):
        u = mda.Universe(PSF, DCD)
        ag = u.select_atoms('name CA')
        return ag

    @pytest.fixture()
    def helanal(self, psf_ca):
        ha = hel.HELANAL(psf_ca, select='resnum 161-187',
                         flatten_single_helix=True)
        return ha.run(start=10, stop=80)

    def test_regression_summary(self, helanal):
        bends = helanal.results.summary['all_bends']
        old_helanal = read_bending_matrix(HELANAL_BENDING_MATRIX_SUBSET)
        assert_almost_equal(np.triu(bends['mean'], 1), old_helanal['Mean'],
                            decimal=1)
        assert_almost_equal(np.triu(bends['sample_sd'], 1), old_helanal['SD'],
                            decimal=1)
        assert_almost_equal(np.triu(bends['abs_dev'], 1), old_helanal['ABDEV'],
                            decimal=1)

    def test_regression_values(self):
        u = mda.Universe(PDB_small)
        ha = hel.HELANAL(u, select='name CA and resnum 161-187',
                         flatten_single_helix=True)
        ha.run()

        for key, value in HELANAL_SINGLE_DATA.items():
            if 'summary' in key:
                data = ha.results[key.split()[0]]
                calculated = [data.mean(), data.std(ddof=1),
                              np.fabs(data-data.mean()).mean()]
            else:
                calculated = ha.results[key][0]

            msg = 'Mismatch between calculated and reference {}'
            assert_almost_equal(calculated, value,
                                decimal=4,
                                err_msg=msg.format(key))

    def test_multiple_selections(self, psf_ca):
        ha = hel.HELANAL(psf_ca, flatten_single_helix=True,
                         select=('resnum 30-40', 'resnum 60-80'))
        ha.run()
        n_frames = len(psf_ca.universe.trajectory)
        assert len(ha.atomgroups) == 2
        assert len(ha.results.summary) == 2
        assert len(ha.results.all_bends) == 2
        assert ha.results.all_bends[0].shape == (n_frames, 8, 8)
        assert ha.results.all_bends[1].shape == (n_frames, 18, 18)

    def test_universe_from_origins(self, helanal):
        u = helanal.universe_from_origins()
        assert isinstance(u, mda.Universe)
        assert len(u.atoms) == len(helanal.atomgroups[0])-2
        assert len(u.trajectory) == 70

    def test_universe_from_origins_except(self, psf_ca):
        ha = hel.HELANAL(psf_ca, select='resnum 161-187')
        with pytest.raises(ValueError, match=r'before universe_from_origins'):
            u = ha.universe_from_origins()

    def test_multiple_atoms_per_residues(self):
        u = mda.Universe(XYZ)
        with pytest.warns(UserWarning) as rec:
            ha = hel.HELANAL(u, select='name H')
        assert len(rec) == 1
        assert 'multiple atoms' in rec[0].message.args[0]

    def test_residue_gaps_split(self, psf_ca):
        sel = 'resid 6:50 or resid 100:130 or resid 132:148'
        with pytest.warns(UserWarning) as rec:
            ha = hel.HELANAL(psf_ca, select=sel).run()
            assert len(ha.atomgroups) == 3
            assert len(ha.atomgroups[0]) == 45
            assert len(ha.atomgroups[1]) == 31
            assert len(ha.atomgroups[2]) == 17
        assert len(rec) == 1
        warnmsg = rec[0].message.args[0]
        assert 'has gaps in the residues' in warnmsg
        assert 'Splitting into 3 helices' in warnmsg

    def test_residue_gaps_no_split(self, psf_ca):
        sel = 'resid 6:50 or resid 100:130 or resid 132:148'
        with pytest.warns(UserWarning) as rec:
            ha = hel.HELANAL(psf_ca, select=sel,
                             split_residue_sequences=False)
            ha.run()
            assert len(ha.atomgroups) == 1
            assert len(ha.atomgroups[0]) == 45+31+17
        assert len(rec) == 1
        warnmsg = rec[0].message.args[0]
        assert 'has gaps in the residues' in warnmsg
        assert 'Splitting into' not in warnmsg

    def test_len_groups_short(self, psf_ca):
        sel = 'resnum 161-168'
        with pytest.warns(UserWarning, match='Fewer than 9 atoms found'):
            ha = hel.HELANAL(psf_ca, select=sel)
            assert len(ha.atomgroups) < 9

    @pytest.mark.parametrize('ref_axis,screw_angles', [
        # input vectors zigzag between [-1, 0, 0] and [1, 0, 0]
        # global axis is z-axis
        ([0, 0, 1], [180, 0]),  # calculated to x-z plane
        ([1, 0, 0], [180, 0]),  # calculated to x-z plane
        ([-1, 0, 0], [0, 180]),  # calculated to x-z plane upside down
        ([0, 1, 0], [90, -90]),  # calculated to y-z plane
        ([0, -1, 0], [-90, 90]),  # calculated to x-z plane upside down
        # calculated to diagonal xy-z plane rotating around
        ([1, 1, 0], [135, -45]),
        ([-1, 1, 0], [45, -135]),
        ([1, -1, 0], [-135, 45]),
        ([-1, -1, 0], [-45, 135]),
        # calculated to diagonal xyz-z plane w/o z contribution
        ([1, 1, 1], [135, -45]),
        ([1, 1, -1], [135, -45]),
        ([1, -1, 1], [-135, 45]),
        ([-1, 1, 1], [45, -135]),
        ([-1, -1, 1], [-45, 135]),
        ([-1, -1, -1], [-45, 135]),
    ])
    def test_helanal_zigzag(self, zigzag, ref_axis, screw_angles):
        ha = hel.HELANAL(zigzag, select="all", ref_axis=ref_axis,
                         flatten_single_helix=True).run()
        assert_almost_equal(ha.results.local_twists, 180, decimal=4)
        assert_almost_equal(ha.results.local_nres_per_turn, 2, decimal=4)
        assert_almost_equal(ha.results.global_axis, [[0, 0, -1]], decimal=4)
        # all 0 vectors
        assert_almost_equal(ha.results.local_axes, 0, decimal=4)
        assert_almost_equal(ha.results.local_bends, 0, decimal=4)
        assert_almost_equal(ha.results.all_bends, 0, decimal=4)
        assert_almost_equal(ha.results.local_heights, 0, decimal=4)
        assert_almost_equal(ha.results.local_helix_directions[0][0::2],
                            [[-1, 0, 0]]*49, decimal=4)
        assert_almost_equal(ha.results.local_helix_directions[0][1::2],
                            [[1, 0, 0]]*49, decimal=4)
        origins = zigzag.atoms.positions[1:-1].copy()
        origins[:, 0] = 0
        assert_almost_equal(ha.results.local_origins[0], origins, decimal=4)
        assert_almost_equal(ha.results.local_screw_angles[0],
                            screw_angles*49, decimal=4)


def test_vector_of_best_fit():
    line = np.random.rand(3)
    unit = line / np.linalg.norm(line)
    points = line*np.arange(1000)[:, np.newaxis]
    noise = np.random.normal(size=(1000, 3))
    data = points + noise

    vector = hel.vector_of_best_fit(data)
    cos = np.dot(vector, unit)
    assert_almost_equal(abs(cos), 1.0, decimal=5)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from numpy.testing import (
    assert_equal,
)
import pytest

import MDAnalysis as mda

from MDAnalysisTests.datafiles import PSF, DCD


# Legacy tests from before 363
@pytest.fixture()
def universe():
    return mda.Universe(PSF, DCD)


@pytest.fixture()
def g(universe):
    return universe.atoms.segments

def test_newSegmentGroup(universe):
    """test that slicing a SegmentGroup returns a new SegmentGroup (Issue 135)"""
    g = universe.atoms.segments
    newg = g[:]
    assert isinstance(newg, mda.core.groups.SegmentGroup)
    assert_equal(len(newg), len(g))


def test_n_atoms(g):
    assert_equal(g.n_atoms, 3341)


def test_n_residues(g):
    assert_equal(g.n_residues, 214)


def test_resids_dim(g):
    assert_equal(len(g.resids), len(g))
    for seg, resids in zip(g, g.resids):
        assert len(resids) == len(seg.residues)
        assert_equal(seg.residues.resids, resids)


def test_resnums_dim(g):
    assert_equal(len(g.resnums), len(g))
    for seg, resnums in zip(g, g.resnums):
        assert len(resnums) == len(seg.residues)
        assert_equal(seg.residues.resnums, resnums)


def test_segids_dim(g):
    assert_equal(len(g.segids), len(g))


def test_set_segids(universe):
    s = universe.select_atoms('all').segments
    s.segids = 'ADK'
    assert_equal(universe.segments.segids, ['ADK'],
                 err_msg="failed to set_segid on segments")


def test_set_segid_updates_(universe):
    g = universe.select_atoms("resid 10:18").segments
    g.segids = 'ADK'
    assert_equal(g.segids, ['ADK'],
                 err_msg="old selection was not changed in place after set_segid")


def test_set_segids_many():
    u = mda.Universe.empty(n_atoms=6, n_residues=2, n_segments=2,
                           atom_resindex=[0, 0, 0, 1, 1, 1], residue_segindex=[0,1])
    u.add_TopologyAttr('segids', ['A', 'B'])

    # universe with 2 segments, A and B

    u.segments.segids = ['X', 'Y']

    assert u.segments[0].segid == 'X'
    assert u.segments[1].segid == 'Y'

    assert len(u.select_atoms('segid A')) == 0
    assert len(u.select_atoms('segid B')) == 0
    assert len(u.select_atoms('segid X')) == 3
    assert len(u.select_atoms('segid Y')) == 3


def test_atom_order(universe):
    assert_equal(universe.segments.atoms.indices,
                 sorted(universe.segments.atoms.indices))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

from MDAnalysisTests.datafiles import (
    TRZ, TRZ_psf,
    waterPSF, waterDCD,
    XYZ_mini,
)
from numpy.testing import assert_almost_equal
import numpy as np
from unittest import mock
from importlib import reload

import MDAnalysis as mda
from MDAnalysis.analysis.hydrogenbonds import (HydrogenBondAutoCorrel as HBAC,
                                               find_hydrogen_donors)


class TestHydrogenBondAutocorrel(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(TRZ_psf, TRZ)

    @pytest.fixture()
    def hydrogens(self, u):
        return u.atoms.select_atoms('name Hn')

    @pytest.fixture()
    def nitrogens(self, u):
        return u.atoms.select_atoms('name N')

    @pytest.fixture()
    def oxygens(self, u):
        return u.atoms.select_atoms('name O')

    # regression tests for different conditions
    def test_continuous(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='continuous',
                     sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution['results'],
            np.array([ 1.        ,  0.92668623,  0.83137828,
                       0.74486804,  0.67741936,  0.60263932],
                     dtype=np.float32)
        )

    def test_continuous_excl(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='continuous',
                     exclusions=(np.arange(len(hydrogens)), np.array(
                         range(len(oxygens)))),
                     sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution['results'],
            np.array([ 1.        ,  0.92668623,  0.83137828,
                       0.74486804,  0.67741936,  0.60263932],
                     dtype=np.float32)
        )


    def test_intermittent(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='intermittent',
                     sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution['results'],
            np.array([ 1.        ,  0.92668623,  0.84310848,
                       0.79325515,  0.76392961,  0.72287393],
                     dtype=np.float32)
        )


    def test_intermittent_timecut(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='intermittent',
                     time_cut=0.01,  # time cut at traj.dt == continuous
                     sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution['results'],
            np.array([ 1.        ,  0.92668623,  0.83137828,
                       0.74486804,  0.67741936,  0.60263932],
                     dtype=np.float32)
        )

    def test_intermittent_excl(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='intermittent',
                     exclusions=(np.arange(len(hydrogens)), np.array(
                         range(len(oxygens)))),
                     sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution['results'],
            np.array([ 1.        ,  0.92668623,  0.84310848,
                       0.79325515,  0.76392961,  0.72287393],
                     dtype=np.float32)
        )

    # For `solve` the test trajectories aren't long enough
    # So spoof the results and check that solver finds solution
    def test_solve_continuous(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='continuous',
                     sample_time=0.06,
        )

        def actual_function_cont(t):
            A1 = 0.75
            A2 = 0.25
            tau1 = 0.5
            tau2 = 0.1
            return A1 * np.exp(-t/tau1) + A2 * np.exp(-t/tau2)
        hbond.solution['time'] = time = np.arange(0, 0.06, 0.001)
        hbond.solution['results'] = actual_function_cont(time)

        hbond.solve()

        assert_almost_equal(
            hbond.solution['fit'],
            np.array([0.75, 0.5, 0.1]),
        )

    def test_solve_intermittent(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='intermittent',
                     sample_time=0.06,
        )

        def actual_function_int(t):
            A1 = 0.33
            A2 = 0.33
            A3 = 0.34
            tau1 = 5
            tau2 = 1
            tau3 = 0.1
            return A1 * np.exp(-t/tau1) + A2 * np.exp(-t/tau2) + A3 * np.exp(-t/tau3)
        hbond.solution['time'] = time = np.arange(0, 6.0, 0.01)
        hbond.solution['results'] = actual_function_int(time)

        hbond.solve()

        assert_almost_equal(
            hbond.solution['fit'],
            np.array([0.33, 0.33, 5, 1, 0.1]),
        )

    # setup errors
    def test_wronglength_DA(self, u, hydrogens, oxygens, nitrogens):
        with pytest.raises(ValueError):
            HBAC(u,
                 hydrogens=hydrogens[:-1],
                 acceptors=oxygens,
                 donors=nitrogens,
                 bond_type='intermittent',
                 exclusions=(np.arange(len(hydrogens)), np.array(
                         range(len(oxygens)))),
                 sample_time=0.06,
        )

    def test_exclusions(self, u, hydrogens, oxygens, nitrogens):
        excl_list = (np.array(range(len(hydrogens))), np.array(
                         range(len(oxygens))))
        excl_list2 = excl_list[0], excl_list[1][:-1]
        with pytest.raises(ValueError):
            HBAC(u,
                 hydrogens=hydrogens,
                 acceptors=oxygens,
                 donors=nitrogens,
                 bond_type='intermittent',
                 exclusions=excl_list2,
                 sample_time=0.06,
        )

    def test_bond_type_VE(self, u, hydrogens, oxygens, nitrogens):
        with pytest.raises(ValueError):
            HBAC(u,
                 hydrogens=hydrogens,
                 acceptors=oxygens,
                 donors=nitrogens,
                 bond_type='marzipan',
                 exclusions=(np.arange(len(hydrogens)), np.array(range(
                     len(oxygens)))),
                 sample_time=0.06,
        )

    def test_solve_before_run_VE(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='continuous',
                     sample_time=0.06,
        )
        with pytest.raises(ValueError):
            hbond.solve()

    @mock.patch('MDAnalysis.coordinates.TRZ.TRZReader._read_frame')
    def test_unslicable_traj_VE(self, mock_read, u, hydrogens, oxygens, nitrogens):
        mock_read.side_effect = TypeError

        with pytest.raises(ValueError):
            HBAC(
                u,
                hydrogens=hydrogens,
                acceptors=oxygens,
                donors=nitrogens,
                bond_type='continuous',
                sample_time=0.06
        )

    def test_repr(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(u,
                     hydrogens=hydrogens,
                     acceptors=oxygens,
                     donors=nitrogens,
                     bond_type='continuous',
                     sample_time=0.06,
        )
        assert isinstance(repr(hbond), str)


def test_find_donors():
    u = mda.Universe(waterPSF, waterDCD)

    H = u.select_atoms('name H*')

    D = find_hydrogen_donors(H)

    assert len(H) == len(D)
    # check each O is bonded to the corresponding H
    for h_atom, o_atom in zip(H, D):
        assert o_atom in h_atom.bonded_atoms


def test_donors_nobonds():
    u = mda.Universe(XYZ_mini)

    with pytest.raises(mda.NoDataError):
        find_hydrogen_donors(u.atoms)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    XPDB_small,
)


class TestXPDBParser(ParserBase):
    parser = mda.topology.ExtendedPDBParser.ExtendedPDBParser
    ref_filename = XPDB_small
    expected_attrs = ['ids', 'names', 'record_types', 'resids',
                      'resnames', 'altLocs', 'icodes', 'occupancies',
                      'tempfactors', 'chainIDs']
    guessed_attrs = ['masses', 'types']
    expected_n_atoms = 5
    expected_n_residues = 5
    expected_n_segments = 1



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from numpy.testing import assert_equal
import numpy as np

import MDAnalysis as mda
from MDAnalysis.topology import guessers
from MDAnalysis.core.topologyattrs import Angles

from MDAnalysisTests import make_Universe
from MDAnalysisTests.core.test_fragments import make_starshape
import MDAnalysis.tests.datafiles as datafiles

from MDAnalysisTests.util import import_not_available


try:
    from rdkit import Chem
    from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges
except ImportError:
    pass

requires_rdkit = pytest.mark.skipif(import_not_available("rdkit"),
                                    reason="requires RDKit")


class TestGuessMasses(object):
    def test_guess_masses(self):
        out = guessers.guess_masses(['C', 'C', 'H'])

        assert isinstance(out, np.ndarray)
        assert_equal(out, np.array([12.011, 12.011, 1.008]))

    def test_guess_masses_warn(self):
        with pytest.warns(UserWarning):
            guessers.guess_masses(['X'])

    def test_guess_masses_miss(self):
        out = guessers.guess_masses(['X', 'Z'])
        assert_equal(out, np.array([0.0, 0.0]))

    @pytest.mark.parametrize('element, value', (('H', 1.008), ('XYZ', 0.0), ))
    def test_get_atom_mass(self, element, value):
        assert guessers.get_atom_mass(element) == value

    def test_guess_atom_mass(self):
        assert guessers.guess_atom_mass('1H') == 1.008


class TestGuessTypes(object):
    # guess_types
    # guess_atom_type
    # guess_atom_element
    def test_guess_types(self):
        out = guessers.guess_types(['MG2+', 'C12'])

        assert isinstance(out, np.ndarray)
        assert_equal(out, np.array(['MG', 'C'], dtype=object))

    def test_guess_atom_element(self):
        assert guessers.guess_atom_element('MG2+') == 'MG'

    def test_guess_atom_element_empty(self):
        assert guessers.guess_atom_element('') == ''

    def test_guess_atom_element_singledigit(self):
        assert guessers.guess_atom_element('1') == '1'

    def test_guess_atom_element_1H(self):
        assert guessers.guess_atom_element('1H') == 'H'
        assert guessers.guess_atom_element('2H') == 'H'
    
    @pytest.mark.parametrize('name, element', (
        ('AO5*', 'O'),
        ('F-', 'F'),
        ('HB1', 'H'),
        ('OC2', 'O'),
        ('1he2', 'H'),
        ('3hg2', 'H'),
        ('OH-', 'O'),
        ('HO', 'H'),
        ('he', 'H'), 
        ('zn', 'ZN'),
        ('Ca2+', 'CA'),
        ('CA', 'C'),
    ))
    def test_guess_element_from_name(self, name, element):
        assert guessers.guess_atom_element(name) == element


def test_guess_charge():
    # this always returns 0.0
    assert guessers.guess_atom_charge('this') == 0.0


def test_guess_bonds_Error():
    u = make_Universe(trajectory=True)
    with pytest.raises(ValueError):
        guessers.guess_bonds(u.atoms[:4], u.atoms.positions[:5])


def test_guess_impropers():
    u = make_starshape()

    ag = u.atoms[:5]

    u.add_TopologyAttr(Angles(guessers.guess_angles(ag.bonds)))

    vals = guessers.guess_improper_dihedrals(ag.angles)
    assert_equal(len(vals), 12)


def bond_sort(arr):
    # sort from low to high, also within a tuple
    # e.g. ([5, 4], [0, 1], [0, 3]) -> ([0, 1], [0, 3], [4, 5])
    out = []
    for (i, j) in arr:
        if i > j:
            i, j = j, i
        out.append((i, j))
    return sorted(out)

def test_guess_bonds_water():
    u = mda.Universe(datafiles.two_water_gro)
    bonds = bond_sort(guessers.guess_bonds(u.atoms, u.atoms.positions, u.dimensions))
    assert_equal(bonds, ((0, 1),
                         (0, 2),
                         (3, 4),
                         (3, 5)))

def test_guess_bonds_adk():
    u = mda.Universe(datafiles.PSF, datafiles.DCD)
    u.atoms.types = guessers.guess_types(u.atoms.names)
    bonds = bond_sort(guessers.guess_bonds(u.atoms, u.atoms.positions))
    assert_equal(np.sort(u.bonds.indices, axis=0),
                 np.sort(bonds, axis=0))

def test_guess_bonds_peptide():
    u = mda.Universe(datafiles.PSF_NAMD, datafiles.PDB_NAMD)
    u.atoms.types = guessers.guess_types(u.atoms.names)
    bonds = bond_sort(guessers.guess_bonds(u.atoms, u.atoms.positions))
    assert_equal(np.sort(u.bonds.indices, axis=0),
                 np.sort(bonds, axis=0))


@pytest.mark.parametrize("smi", [
    "c1ccccc1",
    "C1=CC=CC=C1",
    "CCO",
    "c1ccccc1Cc1ccccc1",
    "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
])
@requires_rdkit
def test_guess_aromaticities(smi):
    mol = Chem.MolFromSmiles(smi)
    mol = Chem.AddHs(mol)
    expected = np.array([atom.GetIsAromatic() for atom in mol.GetAtoms()])
    u = mda.Universe(mol)
    values = guessers.guess_aromaticities(u.atoms)
    assert_equal(values, expected)


@pytest.mark.parametrize("smi", [
    "c1ccccc1",
    "C1=CC=CC=C1",
    "CCO",
    "c1ccccc1Cc1ccccc1",
    "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
])
@requires_rdkit
def test_guess_gasteiger_charges(smi):
    mol = Chem.MolFromSmiles(smi)
    mol = Chem.AddHs(mol)
    ComputeGasteigerCharges(mol, throwOnParamFailure=True)
    expected = np.array([atom.GetDoubleProp("_GasteigerCharge")
                         for atom in mol.GetAtoms()], dtype=np.float32)
    u = mda.Universe(mol)
    values = guessers.guess_gasteiger_charges(u.atoms)
    assert_equal(values, expected)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
import matplotlib
import pytest

import MDAnalysis as mda
from MDAnalysisTests.datafiles import (GRO, XTC, TPR, DihedralArray,
                                       DihedralsArray, RamaArray, GLYRamaArray,
                                       JaninArray, LYSJaninArray, PDB_rama,
                                       PDB_janin)
from MDAnalysis.analysis.dihedrals import Dihedral, Ramachandran, Janin


class TestDihedral(object):

    @pytest.fixture()
    def atomgroup(self):
        u = mda.Universe(GRO, XTC)
        ag = u.select_atoms("(resid 4 and name N CA C) or (resid 5 and name N)")
        return ag


    def test_dihedral(self, atomgroup):
        dihedral = Dihedral([atomgroup]).run()
        test_dihedral = np.load(DihedralArray)

        assert_almost_equal(dihedral.results.angles, test_dihedral, 5,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_dihedral_single_frame(self, atomgroup):
        dihedral = Dihedral([atomgroup]).run(start=5, stop=6)
        test_dihedral = [np.load(DihedralArray)[5]]

        assert_almost_equal(dihedral.results.angles, test_dihedral, 5,
                            err_msg="error: dihedral angles should "
                            "match test vales")

    def test_atomgroup_list(self, atomgroup):
        dihedral = Dihedral([atomgroup, atomgroup]).run()
        test_dihedral = np.load(DihedralsArray)

        assert_almost_equal(dihedral.results.angles, test_dihedral, 5,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_enough_atoms(self, atomgroup):
        with pytest.raises(ValueError):
            dihedral = Dihedral([atomgroup[:2]]).run()

    def test_dihedral_attr_warning(self, atomgroup):
        dihedral = Dihedral([atomgroup]).run(stop=2)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(dihedral.angles, dihedral.results.angles)


class TestRamachandran(object):

    @pytest.fixture()
    def universe(self):
        return mda.Universe(GRO, XTC)

    @pytest.fixture()
    def rama_ref_array(self):
        return np.load(RamaArray)

    def test_ramachandran(self, universe, rama_ref_array):
        rama = Ramachandran(universe.select_atoms("protein")).run()

        assert_almost_equal(rama.results.angles, rama_ref_array, 5,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_ramachandran_single_frame(self, universe, rama_ref_array):
        rama = Ramachandran(universe.select_atoms("protein")).run(
            start=5, stop=6)

        assert_almost_equal(rama.results.angles[0], rama_ref_array[5], 5,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_ramachandran_residue_selections(self, universe):
        rama = Ramachandran(universe.select_atoms("resname GLY")).run()
        test_rama = np.load(GLYRamaArray)

        assert_almost_equal(rama.results.angles, test_rama, 5,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_outside_protein_length(self, universe):
        with pytest.raises(ValueError):
            rama = Ramachandran(universe.select_atoms("resid 220"),
                                check_protein=True).run()

    def test_outside_protein_unchecked(self, universe):
        rama = Ramachandran(universe.select_atoms("resid 220"),
                            check_protein=False).run()

    def test_protein_ends(self, universe):
        with pytest.warns(UserWarning) as record:
            rama = Ramachandran(universe.select_atoms("protein"),
                                check_protein=True).run()
        assert len(record) == 1

    def test_None_removal(self):
        with pytest.warns(UserWarning):
            u = mda.Universe(PDB_rama)
            rama = Ramachandran(u.select_atoms("protein").residues[1:-1])

    def test_plot(self, universe):
        ax = Ramachandran(universe.select_atoms("resid 5-10")).run().plot(ref=True)
        assert isinstance(ax, matplotlib.axes.Axes), \
            "Ramachandran.plot() did not return and Axes instance"

    def test_ramachandran_attr_warning(self, universe):
        rama = Ramachandran(universe.select_atoms("protein")).run(stop=2)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(rama.angles, rama.results.angles)


class TestJanin(object):

    @pytest.fixture()
    def universe(self):
        return mda.Universe(GRO, XTC)

    @pytest.fixture()
    def universe_tpr(self):
        return mda.Universe(TPR, XTC)

    @pytest.fixture()
    def janin_ref_array(self):
        return np.load(JaninArray)

    def test_janin(self, universe, janin_ref_array):
        self._test_janin(universe, janin_ref_array)

    def test_janin_tpr(self, universe_tpr, janin_ref_array):
        """Test that CYSH are filtered (#2898)"""
        self._test_janin(universe_tpr, janin_ref_array)

    def _test_janin(self, u, ref_array):
        janin = Janin(u.select_atoms("protein")).run()

        # Test precision lowered to account for platform differences with osx
        assert_almost_equal(janin.results.angles, ref_array, 3,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_janin_single_frame(self, universe, janin_ref_array):
        janin = Janin(universe.select_atoms("protein")).run(start=5, stop=6)

        assert_almost_equal(janin.results.angles[0], janin_ref_array[5], 3,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_janin_residue_selections(self, universe):
        janin = Janin(universe.select_atoms("resname LYS")).run()
        test_janin = np.load(LYSJaninArray)

        assert_almost_equal(janin.results.angles, test_janin, 3,
                            err_msg="error: dihedral angles should "
                            "match test values")

    def test_outside_protein_length(self, universe):
        with pytest.raises(ValueError):
            janin = Janin(universe.select_atoms("resid 220")).run()

    def test_remove_residues(self, universe):
        with pytest.warns(UserWarning):
            janin = Janin(universe.select_atoms("protein")).run()

    def test_atom_selection(self):
        with pytest.raises(ValueError):
            u = mda.Universe(PDB_janin)
            janin = Janin(u.select_atoms("protein and not resname ALA CYS GLY "
                                         "PRO SER THR VAL"))

    def test_plot(self, universe):
        ax = Janin(universe.select_atoms("resid 5-10")).run().plot(ref=True)
        assert isinstance(ax, matplotlib.axes.Axes), \
            "Ramachandran.plot() did not return and Axes instance"

    def test_janin_attr_warning(self, universe):
        janin = Janin(universe.select_atoms("protein")).run(stop=2)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(janin.angles, janin.results.angles)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import MDAnalysis as mda
from MDAnalysis.core.topology import Topology

mandatory_attrs = ['ids', 'masses', 'types', 
                   'resids', 'resnums', 'segids']


class ParserBase(object):
    """Base class for testing Topology parsers.

    All Parsers must subclass this class!
    """

    expected_attrs = []
    guessed_attrs = []

    @pytest.fixture
    def filename(self):
        return self.ref_filename

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse()

    def test_output(self, filename):
        """Testing the call signature"""
        with self.parser(filename) as p:
            top = p.parse()

        assert isinstance(top, Topology)

    def test_mandatory_attributes(self, top):
        # attributes required as part of the API
        # ALL parsers must provide these
        for attr in mandatory_attrs:
            assert hasattr(top, attr), 'Missing required attribute: {}'.format(attr)

    def test_expected_attributes(self, top):
        # Extra attributes as declared in specific implementations
        for attr in self.expected_attrs+self.guessed_attrs:
            assert hasattr(top, attr), 'Missing expected attribute: {}'.format(attr)
    
    def test_no_unexpected_attributes(self, top):
        attrs = set(self.expected_attrs
                    + self.guessed_attrs
                    + mandatory_attrs
                    + ['indices', 'resindices', 'segindices'])
        for attr in top.attrs:
            assert attr.attrname in attrs, 'Unexpected attribute: {}'.format(attr.attrname)

    def test_guessed_attributes(self, top):
        # guessed attributes must be declared as guessed
        for attr in top.attrs:
            val = attr.is_guessed
            if not val in (True, False):  # only for simple yes/no cases
                continue
            assert val == (attr.attrname in self.guessed_attrs), 'Attr "{}" guessed= {}'.format(attr, val)

    def test_size(self, top):
        """Check that the Topology is correctly sized"""
        assert top.n_atoms == self.expected_n_atoms, '{} atoms read, {} expected in {}'.format(
            top.n_atoms, self.expected_n_atoms, self.__class__.__name__)

        assert top.n_residues == self.expected_n_residues, '{} residues read, {} expected in {}'.format(
            top.n_residues, self.expected_n_residues, self.__class__.__name__)

        assert top.n_segments == self.expected_n_segments, '{} segment read, {} expected in {}'.format(
            top.n_segments, self.expected_n_segments, self.__class__.__name__)

    def test_tt_size(self, top):
        """Check that the transtable is appropriately sized"""
        assert top.tt.size == (self.expected_n_atoms, self.expected_n_residues, self.expected_n_segments)

    def test_creates_universe(self, filename):
        """Check that Universe works with this Parser"""
        u = mda.Universe(filename)
        assert isinstance(u, mda.Universe)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import scipy
import scipy.spatial

import MDAnalysis
from MDAnalysisTests.datafiles import GRO

import MDAnalysis.analysis.distances

from numpy.testing import (assert_equal, assert_array_equal, assert_almost_equal, 
                           assert_array_almost_equal,assert_allclose)
import numpy as np


class TestContactMatrix(object):
    @staticmethod
    @pytest.fixture()
    def coord():
        return np.array([[1, 1, 1],
                      [5, 5, 5],
                      [1.1, 1.1, 1.1],
                      [11, 11, 11],  # neighboring image with pbc
                      [21, 21, 21]],  # non neighboring image with pbc
                     dtype=np.float32)
    
    @staticmethod
    @pytest.fixture()
    def box():
        return np.array([10, 10, 10, 90, 90, 90], dtype=np.float32)
    
    @staticmethod
    @pytest.fixture()
    def shape():
        return 5, 5
    
    @staticmethod
    @pytest.fixture()
    def res_no_pbc():
        return np.array([[1, 0, 1, 0, 0],
                           [0, 1, 0, 0, 0],
                           [1, 0, 1, 0, 0],
                           [0, 0, 0, 1, 0],
                           [0, 0, 0, 0, 1]], dtype=bool)
    
    @staticmethod
    @pytest.fixture()
    def res_pbc():
        return np.array([[1, 0, 1, 1, 1],
                        [0, 1, 0, 0, 0],
                        [1, 0, 1, 1, 1],
                        [1, 0, 1, 1, 1],
                        [1, 0, 1, 1, 1]], dtype=bool)

    def test_np(self, coord, shape, res_no_pbc):
        contacts = MDAnalysis.analysis.distances.contact_matrix(
            coord, cutoff=1, returntype="numpy"
        )
        assert contacts.shape == shape, \
            "wrong shape (should be {0})".format(shape)
        assert_equal(contacts, res_no_pbc)

    def test_sparse(self, coord, shape, res_no_pbc):
        contacts = MDAnalysis.analysis.distances.contact_matrix(
            coord, cutoff=1.5, returntype="sparse"
        )
        assert contacts.shape == shape, \
            "wrong shape (should be {0})".format(shape)
        assert_equal(contacts.toarray(), res_no_pbc)

    def test_box_numpy(self, coord, box, shape, res_pbc):
        contacts = MDAnalysis.analysis.distances.contact_matrix(
            coord, box=box, cutoff=1
        )
        assert contacts.shape == shape, \
            "wrong shape (should be {0})".format(shape)
        assert_equal(contacts, res_pbc)

    def test_box_sparse(self, coord, box, shape, res_pbc):
        contacts = MDAnalysis.analysis.distances.contact_matrix(
            coord, box=box, cutoff=1, returntype='sparse'
        )
        assert contacts.shape == shape, \
            "wrong shape (should be {0})".format(shape)
        assert_equal(contacts.toarray(), res_pbc)



class TestDist(object):

    @staticmethod
    @pytest.fixture()
    def ag():
        u = MDAnalysis.Universe(GRO)
        return u.atoms[1:10]

    # TODO: How are ag and ag2 different?!
    @staticmethod
    @pytest.fixture()
    def ag2():
        u2 = MDAnalysis.Universe(GRO)
        return u2.atoms[11:20]

    @staticmethod
    @pytest.fixture()
    def box():
        return np.array([8, 8, 8, 90, 90, 90], dtype=np.float32)
    
    @staticmethod
    @pytest.fixture()
    def expected(ag, ag2):
        
        return np.diag(scipy.spatial.distance.cdist(
            ag.positions, ag2.positions)
        )

    @staticmethod
    @pytest.fixture()
    def expected_box(ag, ag2, box):

        rp = np.abs(ag.positions - ag2.positions)
        box_2d = box[np.newaxis, 0:3]
        rp = np.where(rp  > box_2d / 2, box_2d - rp, rp)
        return np.sqrt(np.square(rp).sum(axis=1))

    def test_pairwise_dist(self, ag, ag2, expected):
        '''Ensure that pairwise distances between atoms are
        correctly calculated.'''
        actual = MDAnalysis.analysis.distances.dist(ag, ag2)[2]
        assert_equal(actual, expected)

    def test_pairwise_dist_box(self, ag, ag2, expected_box, box):
        '''Ensure that pairwise distances between atoms are
        correctly calculated.'''
        actual = MDAnalysis.analysis.distances.dist(ag, ag2, 0, box)[2]
        assert_allclose(actual, expected_box, rtol=1e-05, atol=10)

    def test_pairwise_dist_offset_effect(self, ag, ag2, expected):
        '''Test that feeding in offsets to dist() doesn't alter
        pairwise distance matrix.'''
        actual = MDAnalysis.analysis.distances.dist(
            ag, ag2, offset=229)[2]
        assert_equal(actual, expected)

    def test_offset_calculation(self, ag, ag2):
        '''Test that offsets fed to dist() are correctly calculated.'''
        actual = MDAnalysis.analysis.distances.dist(ag, ag2,
                                                    offset=33)[:2]
        assert_equal(actual, np.array([ag.atoms.resids + 33,
                                       ag2.atoms.resids + 33]))

    def test_mismatch_exception(self, ag, ag2, expected):
        '''A ValueError should be raised if the two atomgroups
        don't have the same number of atoms.'''
        with pytest.raises(ValueError):
            MDAnalysis.analysis.distances.dist(ag[:8], ag2)


class TestBetween(object):
    @staticmethod
    @pytest.fixture()
    def u():
        return MDAnalysis.Universe(GRO)

    @staticmethod
    @pytest.fixture()
    def ag(u):
        return u.atoms[:10]

    @staticmethod
    @pytest.fixture()
    def ag2(u):
        return u.atoms[12:33]

    @staticmethod
    @pytest.fixture()
    def group(u):
        return u.atoms[40:]

    distance = 5.9

    @pytest.fixture()
    def expected(self, group, ag, ag2):
        distance_matrix_1 = scipy.spatial.distance.cdist(group.positions,
                                                         ag.positions)
        mask_1 = np.unique(np.where(distance_matrix_1 <= self.distance)[0])
        group_filtered = group[mask_1]
        distance_matrix_2 = scipy.spatial.distance.cdist(group_filtered.positions,
                                                         ag2.positions)
        mask_2 = np.unique(np.where(distance_matrix_2 <= self.distance)[0])
        return group_filtered[mask_2].indices

    def test_between_simple_case_indices_only(self, group, ag, ag2, expected):
        '''Test MDAnalysis.analysis.distances.between() for
        a simple input case. Checks the sorted atom indices
        of returned AtomGroup against sorted expected index
        values.'''
        actual = sorted(MDAnalysis.analysis.distances.between(
            group,
            ag,
            ag2,
            self.distance
        ).indices)
        assert_equal(actual, expected)



import pytest
from numpy.testing import assert_equal
import mmtf
from unittest import mock

import MDAnalysis as mda
from MDAnalysis.core.groups import AtomGroup

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import MMTF, MMTF_gz, MMTF_skinny, MMTF_skinny2

class MMTFBase(ParserBase):
    expected_attrs = [
        'ids', 'names', 'types', 'altLocs', 'tempfactors', 'occupancies',
        'charges', 'names', 'resnames', 'resids', 'resnums', 'icodes',
        'segids', 'bonds', 'models'
    ]


class TestMMTFParser(MMTFBase):
    parser = mda.topology.MMTFParser.MMTFParser
    ref_filename = MMTF
    guessed_attrs = ['masses']
    expected_n_atoms = 512
    expected_n_residues = 124
    expected_n_segments = 8


class TestMMTFParser_gz(TestMMTFParser):
    ref_filename = MMTF_gz
    expected_n_atoms = 1140
    expected_n_residues = 36
    expected_n_segments = 4


class TestMMTFSkinny(MMTFBase):
    parser = mda.topology.MMTFParser.MMTFParser
    ref_filename = MMTF_skinny
    # for all attributes often in MMTF,
    # check that we get expected error on access
    # (sort so pytest gets reliable order)
    guessed_attrs = ['ids', 'masses', 'segids']
    expected_n_atoms = 660
    expected_n_residues = 134
    expected_n_segments = 2

    
class TestMMTFSkinny2(MMTFBase):
    parser = mda.topology.MMTFParser.MMTFParser
    ref_filename = MMTF_skinny2
    guessed_attrs = ['ids', 'masses', 'segids']
    expected_n_atoms = 169
    expected_n_residues = 44
    expected_n_segments = 2


class TestMMTFUniverse(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(MMTF)

    def test_bonds(self, u):
        assert len(u.bonds) == 458

    def test_ids(self, u):
        assert_equal(u.atoms.ids[:3], [1, 2, 3])

    def test_names(self, u):
        assert_equal(u.atoms.names[:3], ["O5'", "C5'", "C4'"])

    def test_resnames(self, u):
        assert_equal(u.residues.resnames[:3], ['DG', 'DA', 'DA'])

    def test_segids(self, u):
        assert_equal(u.segments[:3].segids, ['A', 'B', 'C'])

    def test_resids(self, u):
        assert_equal(u.residues.resids[-3:], [2008, 2009, 2010])

    def test_occupancies(self, u):
        # pylint: disable=unsubscriptable-object
        assert_equal(u.atoms.occupancies[:3], [1.0, 1.0, 1.0])

    def test_bfactors(self, u):
        assert_equal(u.atoms.bfactors[:3], [9.48, 10.88, 10.88])

    def test_types(self, u):
        assert_equal(u.atoms.types[:3], ['O', 'C', 'C'])

    def test_models(self, u):
        assert all(u.atoms.models == 0)

    def test_icodes(self, u):
        assert all(u.atoms.icodes == '')

    def test_altlocs(self, u):
        assert all(u.atoms.altLocs[:3] == '')


class TestMMTFUniverseFromDecoder(TestMMTFUniverse):
    @pytest.fixture()
    def u(self):
        return mda.Universe(mmtf.parse(MMTF))


class TestMMTFgzUniverse(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(MMTF_gz)

    def test_models(self, u):
        # has 2 models
        assert_equal(u.segments[:2].models, [0, 0])
        assert_equal(u.segments[2:].models, [1, 1])

    def test_universe_models(self, u):
        assert len(u.models) == 2
        for m in u.models:
            assert isinstance(m, AtomGroup)
            assert len(m) == 570


class TestMMTFgzUniverseFromDecoder(TestMMTFgzUniverse):
    @pytest.fixture()
    def u(self):
        top = mmtf.parse_gzip(MMTF_gz)
        return mda.Universe(top)


class TestMMTFFetch(TestMMTFUniverse):
    @pytest.fixture()
    def u(self):
        top = mmtf.parse(MMTF)
        with mock.patch('mmtf.fetch') as mock_fetch:
            mock_fetch.return_value = top

            return mda.fetch_mmtf('173D')  # string is irrelevant


class TestSelectModels(object):
    # tests for 'model' keyword in select_atoms   
    @pytest.fixture()
    def u(self):
        return mda.Universe(MMTF_gz)

    def test_model_selection(self, u):
        m1 = u.select_atoms('model 0')
        m2 = u.select_atoms('model 1')

        assert len(m1) == 570
        assert len(m2) == 570

    def test_model_multiple(self, u):
        m2plus = u.select_atoms('model 1-10')

        assert len(m2plus) == 570

    def test_model_multiple_2(self, u):
        m2plus = u.select_atoms('model 1:10')

        assert len(m2plus) == 570

    def test_model_multiple_3(self, u):
        m1and2 = u.select_atoms('model 0-1')

        assert len(m1and2) == 1140

    def test_model_multiple_4(self, u):
        m1and2 = u.select_atoms('model 0:1')

        assert len(m1and2) == 1140

    def test_model_multiple_5(self, u):
        m1and2 = u.select_atoms('model 0 1')

        assert len(m1and2) == 1140



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""32 bit compat tests

Tests for making sure that integer arrays used for indexing use `np.intp`.
This dtype is important for platform independent indexing of other arrays.

"""
import numpy as np
import pytest
from MDAnalysisTests import make_Universe


@pytest.fixture()
def u():
    return make_Universe()


def test_ag_ix(u):
    assert u.atoms.ix.dtype == np.intp


def test_rg_ix(u):
    assert u.residues.ix.dtype == np.intp


def test_sg_ix(u):
    assert u.segments.ix.dtype == np.intp


def test_atom_ix_array(u):
    assert u.atoms[0].ix_array.dtype == np.intp


def test_residue_ix_array(u):
    assert u.residues[0].ix_array.dtype == np.intp


def test_segment_ix_array(u):
    assert u.segments[0].ix_array.dtype == np.intp


def test_atomgroup_indices(u):
    assert u.atoms.indices.dtype == np.intp


def test_atomgroup_residue_upshift(u):
    assert u.atoms.resindices.dtype == np.intp


def test_atomgroup_segment_upshift(u):
    assert u.atoms.segindices.dtype == np.intp


def test_residuegroup_atom_downshift(u):
    # downshift arrays are a list (one for each residue)
    assert all((arr.dtype == np.intp)
               for arr in u.residues.indices)


def test_residuegroup_resindices(u):
    assert u.residues.resindices.dtype == np.intp


def test_residuegroup_segment_upshift(u):
    assert u.residues.segindices.dtype == np.intp


def test_segmentgroup_atom_downshift(u):
    assert all((arr.dtype == np.intp)
               for arr in u.segments.indices)


def test_segmentgroup_residue_downshift(u):
    assert all((arr.dtype == np.intp)
               for arr in u.segments.resindices)


def test_segmentgroup_segindices(u):
    assert u.segments.segindices.dtype == np.intp
